import json
from typing import Any, Dict, List, Optional, Union

from openai.types.chat.chat_completion_message import (
    ChatCompletionMessageToolCall,
    FunctionCall,
)
from pydantic import BaseModel, Field


class Message(BaseModel):
    """Chat message model compatible with OpenAI's interface."""

    role: str
    content: Optional[str] = ""  # Content can be None for tool calls in OpenAI API
    name: Optional[str] = None
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None
    function_call: Optional[FunctionCall] = None

    @classmethod
    def model_validate(cls, obj, *args, **kwargs):
        if isinstance(obj, dict) and "role" not in obj:
            raise ValueError("Role is required")
        return super().model_validate(obj, *args, **kwargs)


class MetricResult(BaseModel):
    """Result of a single metric evaluation.

    Attributes:
        is_score_valid (bool): Whether the score is valid for this metric (required).
        score (float): The score for this metric.
        reason (str): Explanation for the score.
    """

    is_score_valid: bool = True
    score: float = Field(..., ge=0.0, le=1.0)
    reason: str

    def __getitem__(self, key: str) -> Any:
        if (
            key in self.__fields__
        ):  # Changed to __fields__ for Pydantic v1 compatibility
            value = getattr(self, key)
            return value
        raise KeyError(f"'{key}'")

    def __contains__(self, key: str) -> bool:
        return key in self.__fields__  # Changed to __fields__

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def keys(self):
        return self.__fields__.keys()  # Changed to __fields__

    def values(self):
        # For consistency with __getitem__ returning raw attribute values (including nested models)
        return [
            getattr(self, key) for key in self.__fields__.keys()
        ]  # Changed to __fields__

    def items(self):
        return [
            (key, getattr(self, key)) for key in self.__fields__.keys()
        ]  # Changed to __fields__

    def __iter__(self):
        return iter(self.__fields__.keys())  # Changed to __fields__


class StepOutput(BaseModel):
    """Defines the base reward and other metrics for a single conceptual step within a rollout,
    as determined by the user's reward function.
    """

    step_index: Union[int, str] = Field(
        description="User-defined index for the step (e.g., assistant message index, turn number). This is used by the system to map this output to the internal StepData."
    )
    base_reward: float = Field(
        description="Base reward calculated by the user's reward function for this step."
    )
    metrics: Dict[str, Any] = Field(
        default_factory=dict,
        description="Optional dictionary of custom metrics for this step.",
    )
    reason: Optional[str] = Field(
        default=None,
        description="Optional explanation for the step's base reward or metrics.",
    )


class EvaluateResult(BaseModel):
    """The complete result of an evaluator.
    For standard evaluation, it provides an overall score and component metrics.
    For Reinforcement Learning, it can also provide per-step base rewards via 'step_outputs'.

    Attributes:
        score (float): The overall evaluation score.
        is_score_valid (bool): Whether the overall score is valid. Defaults to True.
        reason (Optional[str]): Optional explanation for the overall score.
        metrics (Dict[str, MetricResult]): Dictionary of component metrics for detailed evaluation.
        step_outputs (Optional[List[StepOutput]]): For RL, a list of outputs for each conceptual step,
                                                  providing base rewards.
        error (Optional[str]): Optional error message if evaluation failed.
    """

    score: float = Field(
        ..., description="The overall evaluation score, typically between 0.0 and 1.0."
    )
    is_score_valid: bool = Field(
        default=True, description="Whether the overall score is valid."
    )
    reason: Optional[str] = Field(
        default=None, description="Optional explanation for the overall score."
    )
    metrics: Dict[str, MetricResult] = Field(
        default_factory=dict,
        description="Dictionary of component metrics for detailed breakdown.",
    )

    # New field for RL per-step base rewards
    step_outputs: Optional[List[StepOutput]] = Field(
        default=None,
        description="For RL, a list of outputs for each conceptual step, providing base rewards.",
    )

    error: Optional[str] = Field(
        default=None,
        description="Optional error message if the evaluation itself encountered an issue.",
    )

    def __getitem__(self, key: str) -> Any:
        if key in self.__fields__:  # Changed to __fields__
            value = getattr(self, key)
            # If the value is a dict of MetricResult, and we want __getitem__ on metrics
            # to return a dict of dicts (rather than dict of MetricResult objects),
            # we'd need special handling here.
            # For now, return the raw attribute value, consistent with MetricResult.__getitem__
            return value
        raise KeyError(f"'{key}'")

    def __contains__(self, key: str) -> bool:
        return key in self.__fields__  # Changed to __fields__

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def keys(self):
        return self.__fields__.keys()  # Changed to __fields__

    def values(self):
        # For consistency with __getitem__ returning raw attribute values
        return [
            getattr(self, key) for key in self.__fields__.keys()
        ]  # Changed to __fields__

    def items(self):
        return [
            (key, getattr(self, key)) for key in self.__fields__.keys()
        ]  # Changed to __fields__

    def __iter__(self):
        return iter(self.__fields__.keys())  # Changed to __fields__


# Original dataclass-based models for backwards compatibility
# These are deprecated and will be removed in a future version
# Use EvaluateResult and MetricResult instead
# MetricRewardOutput and RewardOutput are fully removed.


# --- Models for New Agent Evaluation Framework (V2) ---


class ResourceServerConfig(BaseModel):
    """
    Configuration for a resource server required by a task.
    """

    start_command: str = Field(
        description="The command to start the server. The string '{port}' will be replaced with a dynamically allocated free port."
    )
    health_check_url: str = Field(
        description="The URL to poll to check if the server is ready. The string '{port}' will be replaced with the allocated port."
    )


class EvaluationCriteriaModel(BaseModel):
    """
    Defines criteria for evaluating task success, often by querying the final state of a resource.
    """

    final_state_query: Optional[str] = Field(
        default=None,
        description="A query (e.g., SQL) to run on the final state of the resource.",
    )
    expected_query_result_transform: Optional[str] = Field(
        default=None,
        description="A Python lambda string (e.g., 'lambda x: x > 0') to transform and evaluate the query result to a boolean.",
    )

    # Explicit fields for ground truth data for BFCL evaluation
    ground_truth_function_calls: Optional[List[List[str]]] = Field(
        default=None, description="Ground truth function calls for BFCL evaluation."
    )
    ground_truth_comparable_state: Optional[Dict[str, Any]] = Field(
        default=None, description="Ground truth comparable state for BFCL evaluation."
    )

    # Future: Could include other complex evaluation logic or references


class TaskDefinitionModel(BaseModel):
    """
    Pydantic model for validating the structure of a V2 agent evaluation task definition file (YAML/JSON).
    """

    name: str = Field(description="Unique name for the task.")
    description: Optional[str] = Field(
        default=None, description="A brief description of the task."
    )

    resource_type: str = Field(
        description="The type of ForkableResource to use (e.g., 'SQLResource', 'PythonStateResource', 'FileSystemResource', 'DockerResource')."
    )
    base_resource_config: Dict[str, Any] = Field(
        default_factory=dict,
        description="Configuration dictionary passed to the base resource's setup() method.",
    )

    tools_module_path: Optional[str] = Field(
        default=None,
        description="Optional Python import path to a module containing custom tool functions for this task.",
    )
    reward_function_path: str = Field(
        description="Python import path to the reward function (e.g., 'my_module.my_reward_func')."
    )

    goal_description: Optional[str] = Field(
        default=None,
        description="A human-readable description of the agent's goal for this task.",
    )
    evaluation_criteria: Optional[EvaluationCriteriaModel] = Field(
        default=None,
        description="Criteria used by the Orchestrator to determine if the primary goal was achieved.",
    )

    initial_user_prompt: Optional[str] = Field(
        default=None,
        description="The initial prompt or message to start the agent interaction. Deprecated if 'messages' field is used for multi-turn.",
    )
    messages: Optional[List[Dict[str, Any]]] = (
        Field(  # Explicit field for initial/multi-turn messages
            default=None,
            description="A list of messages to start the conversation, can represent multiple user turns for sequential processing.",
        )
    )

    # PoC / Task specific parameters
    poc_max_turns: int = Field(
        default=3,
        ge=1,
        description="For PoC Orchestrator, the maximum number of interaction turns.",
    )

    # Allow other custom fields to be captured if needed by specific tasks or resources
    # These will be accessible via `model_extra` if `model_config` has `extra = 'allow'`
    # Or define a specific field:
    # custom_task_params: Dict[str, Any] = Field(default_factory=dict)
    resource_server: Optional[ResourceServerConfig] = Field(
        default=None,
        description="Configuration for a background server required for the task.",
    )

    num_rollouts: int = Field(
        default=1,
        ge=1,
        description="Number of parallel rollouts to execute for this task definition.",
    )

    # Data-driven evaluation fields
    dataset_path: Optional[str] = Field(
        default=None,
        description="Path to dataset file (JSONL) containing experimental conditions for data-driven evaluation.",
    )
    num_rollouts_per_sample: int = Field(
        default=1,
        ge=1,
        description="Number of rollouts to execute per sample from the dataset.",
    )

    class Config:
        extra = "allow"  # Allow and capture extra fields not explicitly defined
        # For Pydantic v2, it's model_config = {"extra": "allow"}
        # Assuming Pydantic v1 style for now based on existing file, can update if needed.
        # If using Pydantic v2, this should be:
        # from pydantic import ConfigDict
        # model_config = ConfigDict(extra='allow')
        # For Pydantic v1, `Config.extra = "allow"` is correct.



"""
Tool Registry for the Agent Evaluation Framework.
Provides a mechanism to register and manage tools.
"""

import functools
import inspect
from typing import Any, Callable, Dict, List, Optional


class ToolRegistry:
    """
    Registry for tools that can be used by agents.

    Attributes:
        name: Unique identifier for this tool registry
        tools: Dictionary mapping tool names to tool functions
        descriptions: Dictionary mapping tool names to their descriptions
        parameters: Dictionary mapping tool names to their parameter specifications
    """

    def __init__(self, name: str):
        """
        Initialize a new tool registry.

        Args:
            name: Unique identifier for this registry
        """
        self.name = name
        self.tools: Dict[str, Callable] = {}
        self.descriptions: Dict[str, str] = {}
        self.parameters: Dict[str, Dict[str, Any]] = {}

    def tool(self, description: str, parameters: Dict[str, Any]) -> Callable:
        """
        Decorator to register a function as a tool.

        Args:
            description: Human-readable description of the tool
            parameters: Parameter specifications for the tool

        Returns:
            Decorator function that registers the decorated function
        """

        def decorator(func: Callable) -> Callable:
            tool_name = func.__name__
            self.tools[tool_name] = func
            self.descriptions[tool_name] = description
            self.parameters[tool_name] = parameters

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def get_tool(self, tool_name: str) -> Optional[Callable]:
        """
        Get a tool function by name.

        Args:
            tool_name: Name of the tool to retrieve

        Returns:
            The tool function if found, None otherwise
        """
        return self.tools.get(tool_name)

    def get_tools(self) -> Dict[str, Callable]:
        """
        Get all tools in this registry.

        Returns:
            Dictionary mapping tool names to tool functions
        """
        return self.tools

    def get_openai_tools(self) -> List[Dict[str, Any]]:
        """
        Get tool specifications in OpenAI function calling format.

        Returns:
            List of tool specifications compatible with OpenAI API
        """
        tools = []
        for tool_name in self.tools:
            tools.append(
                {
                    "name": tool_name,
                    "description": self.descriptions.get(tool_name, ""),
                    "parameters": {
                        "type": "object",
                        "properties": self.parameters.get(tool_name, {}),
                        "required": list(self.parameters.get(tool_name, {}).keys()),
                    },
                }
            )
        return tools

    def create_fastapi_app(self):
        """
        Create a FastAPI app with endpoints for each tool.

        Returns:
            A FastAPI app instance with tool endpoints
        """
        # This is a stub implementation
        return {"app_type": "FastAPI", "tools": list(self.tools.keys())}



"""
End-to-end integration tests for batch evaluation feature.

These tests validate the entire batch evaluation pipeline with live API calls
to both Fireworks and OpenAI, ensuring production readiness.
"""

import asyncio
import json
import logging
import os
import tempfile
from pathlib import Path
from typing import Any, Dict
from unittest.mock import AsyncMock, Mock, patch

import pytest

from reward_kit.agent.task_manager import TaskManager
from reward_kit.cli_commands.agent_eval_cmd import agent_eval_command
from reward_kit.models import TaskDefinitionModel


class MockArgs:
    """Mock args object for agent_eval_command."""

    def __init__(self, task_def: str, num_rollouts: int = 2, **kwargs):
        self.task_def = task_def
        self.num_rollouts = num_rollouts
        self.parallel = kwargs.get("parallel", False)
        self.max_concurrency = kwargs.get("max_concurrency", 3)
        self.model = kwargs.get("model", None)
        self.filter = kwargs.get("filter", None)


@pytest.mark.integration
class TestBatchEvaluation:
    """Integration tests for batch evaluation functionality."""

    def _create_sophisticated_game_mock(self):
        """Create the sophisticated FrozenLakeGameMock for realistic game simulation."""

        class FrozenLakeGameMock:
            def __init__(self):
                self.episodes = {}
                self.call_count = 0

            def handle_request(self, *args, **kwargs):
                """Handle HTTP requests and simulate Frozen Lake game logic"""
                self.call_count += 1

                # Parse the request to determine the endpoint
                if hasattr(args[0], "endswith"):
                    url = args[0]
                elif "url" in kwargs:
                    url = kwargs["url"]
                else:
                    url = str(args[0]) if args else ""

                # Check if this is a step request with JSON data
                json_data = kwargs.get("json", {})

                if "/start_episode" in url:
                    return self._start_episode()
                elif "/step" in url and json_data:
                    episode_id = json_data.get("episode_id")
                    action = json_data.get("action")
                    return self._step(episode_id, action)
                else:
                    # Default response for other requests
                    return self._default_response()

            def _start_episode(self):
                """Start a new episode"""
                episode_id = f"episode_{self.call_count}"
                self.episodes[episode_id] = {
                    "position": [0, 0],  # Start position
                    "step_count": 0,
                    "done": False,
                    "won": False,
                    "grid": [
                        ["S", "F", "F", "F"],
                        ["F", "H", "F", "H"],
                        ["F", "F", "F", "H"],
                        ["H", "F", "F", "G"],
                    ],
                }

                observation = {
                    "position": [0, 0],
                    "current_cell": "S",
                    "done": False,
                    "won": False,
                    "message": "Game started. You are at the starting position.",
                    "visual": self._generate_visual(episode_id),
                    "step_count": 0,
                }

                response = Mock()
                response.status_code = 200
                response.json.return_value = {
                    "episode_id": episode_id,
                    "observation": observation,
                }
                return response

            def _step(self, episode_id, action):
                """Process a step action"""
                if episode_id not in self.episodes:
                    # Episode doesn't exist, create a basic one
                    self.episodes[episode_id] = {
                        "position": [0, 0],
                        "step_count": 0,
                        "done": False,
                        "won": False,
                        "grid": [
                            ["S", "F", "F", "F"],
                            ["F", "H", "F", "H"],
                            ["F", "F", "F", "H"],
                            ["H", "F", "F", "G"],
                        ],
                    }

                episode = self.episodes[episode_id]

                if episode["done"]:
                    # Episode already finished
                    observation = self._get_observation(episode_id)
                else:
                    # Process the action
                    episode["step_count"] += 1
                    old_pos = episode["position"].copy()
                    new_pos = self._apply_action(episode["position"], action)
                    episode["position"] = new_pos

                    # Check what happened
                    row, col = new_pos
                    cell = episode["grid"][row][col]

                    if cell == "G":
                        # Reached goal!
                        episode["done"] = True
                        episode["won"] = True
                        message = (
                            "Congratulations! You reached the goal! You win! Success!"
                        )
                    elif cell == "H":
                        # Fell in hole
                        episode["done"] = True
                        episode["won"] = False
                        message = "Oh no! You fell into a hole. Game over."
                    else:
                        # Normal move
                        action_names = {0: "left", 1: "down", 2: "right", 3: "up"}
                        action_name = action_names.get(action, "unknown")
                        if new_pos != old_pos:
                            message = f"You moved {action_name} to a {cell} cell."
                        else:
                            message = f"You tried to move {action_name} but hit a wall."

                    observation = {
                        "position": new_pos,
                        "current_cell": cell,
                        "done": episode["done"],
                        "won": episode["won"],
                        "message": message,
                        "visual": self._generate_visual(episode_id),
                        "step_count": episode["step_count"],
                    }

                response = Mock()
                response.status_code = 200
                response.json.return_value = {
                    "observation": observation,
                    "is_done": episode["done"],
                    "info": {"step_count": episode["step_count"]},
                }
                return response

            def _apply_action(self, position, action):
                """Apply action and return new position"""
                row, col = position

                # Action mapping: 0=left, 1=down, 2=right, 3=up
                if action == 0:  # left
                    new_col = max(0, col - 1)
                    return [row, new_col]
                elif action == 1:  # down
                    new_row = min(3, row + 1)
                    return [new_row, col]
                elif action == 2:  # right
                    new_col = min(3, col + 1)
                    return [row, new_col]
                elif action == 3:  # up
                    new_row = max(0, row - 1)
                    return [new_row, col]
                else:
                    return position  # Invalid action, stay in place

            def _generate_visual(self, episode_id):
                """Generate visual representation of the game"""
                episode = self.episodes[episode_id]
                grid = episode["grid"]
                pos = episode["position"]

                visual_lines = []
                for r in range(4):
                    line = ""
                    for c in range(4):
                        if [r, c] == pos:
                            line += f"[{grid[r][c]}] "
                        else:
                            line += f" {grid[r][c]}  "
                    visual_lines.append(line.rstrip())

                return "\n".join(visual_lines)

            def _get_observation(self, episode_id):
                """Get current observation for an episode"""
                episode = self.episodes[episode_id]
                row, col = episode["position"]
                cell = episode["grid"][row][col]

                return {
                    "position": episode["position"],
                    "current_cell": cell,
                    "done": episode["done"],
                    "won": episode["won"],
                    "message": (
                        "Congratulations! You reached the goal! You win! Success!"
                        if episode["won"]
                        else "Game finished."
                    ),
                    "visual": self._generate_visual(episode_id),
                    "step_count": episode["step_count"],
                }

            def _default_response(self):
                """Default response for unhandled requests"""
                response = Mock()
                response.status_code = 200
                response.json.return_value = {"status": "ok"}
                return response

        return FrozenLakeGameMock()

    def setup_method(self):
        """Set up test environment before each test."""
        # Ensure we have the necessary environment variables
        self.original_env = {}

        # Store original environment values
        env_vars = ["FIREWORKS_API_KEY", "OPENAI_API_KEY", "MODEL_AGENT"]
        for var in env_vars:
            self.original_env[var] = os.environ.get(var)

        # Set default model for agent if not specified
        if not os.environ.get("MODEL_AGENT"):
            os.environ["MODEL_AGENT"] = "accounts/fireworks/models/qwen3-235b-a22b"

        # Set mock API keys to avoid skipping tests
        os.environ["FIREWORKS_API_KEY"] = "mock-fireworks-key"
        os.environ["OPENAI_API_KEY"] = "mock-openai-key"

    def teardown_method(self):
        """Clean up after each test."""
        # Restore original environment
        for var, value in self.original_env.items():
            if value is None:
                os.environ.pop(var, None)
            else:
                os.environ[var] = value

    @pytest.mark.asyncio
    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("aiohttp.ClientSession.post")
    @patch("httpx.Client.post")
    @patch("requests.get")
    async def test_batch_evaluation_task_manager_fireworks(
        self,
        mock_requests_get,
        mock_httpx_post,
        mock_aiohttp_post,
        mock_subprocess_popen,
        mock_openai_constructor,
    ):
        """Test batch evaluation using TaskManager with Fireworks API."""
        # Mock OpenAI client in orchestrator
        mock_openai_client = AsyncMock()
        mock_openai_constructor.return_value = mock_openai_client

        # Mock OpenAI completion response - simulate smart AI moves
        def create_tool_call_response(action, call_id):
            mock_tool_call = Mock()
            mock_tool_call.function.name = "step"
            mock_tool_call.function.arguments = f'{{"action": "{action}"}}'
            mock_tool_call.id = call_id

            mock_message = Mock()
            mock_message.content = None
            mock_message.role = "assistant"
            mock_message.tool_calls = [mock_tool_call]
            mock_message.model_dump = Mock(
                return_value={
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": call_id,
                            "type": "function",
                            "function": {
                                "name": "step",
                                "arguments": f'{{"action": "{action}"}}',
                            },
                        }
                    ],
                }
            )

            mock_completion = AsyncMock()
            mock_completion.choices = [Mock(message=mock_message)]
            mock_completion.usage = Mock(total_tokens=10)
            return mock_completion

        # Smart AI with winning sequence: right -> right -> down -> down -> down -> right
        winning_sequence = ["right", "right", "down", "down", "down", "right"]
        rollout_counter = [0]

        def smart_move_generator(**kwargs):
            messages = kwargs.get("messages", [])
            move_count = sum(
                1
                for msg in messages
                if msg.get("role") == "assistant" and msg.get("tool_calls")
            )

            rollout_counter[0] += 1
            if rollout_counter[0] <= 6:  # First rollout wins
                action = (
                    winning_sequence[move_count]
                    if move_count < len(winning_sequence)
                    else "right"
                )
            else:  # Second rollout makes mistake
                action = (
                    "right"
                    if move_count == 4
                    else (
                        winning_sequence[move_count]
                        if move_count < len(winning_sequence)
                        else "right"
                    )
                )

            return create_tool_call_response(action, f"call_{move_count}")

        mock_openai_client.chat.completions.create = AsyncMock(
            side_effect=smart_move_generator
        )

        # Mock Fireworks API response (backup)
        mock_response = AsyncMock()
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value={
                "choices": [
                    {
                        "message": {"content": "right", "role": "assistant"},
                        "finish_reason": "stop",
                    }
                ],
                "usage": {"total_tokens": 10},
            }
        )
        mock_aiohttp_post.return_value.__aenter__.return_value = mock_response

        # Use sophisticated game mock
        game_mock = self._create_sophisticated_game_mock()
        mock_httpx_post.side_effect = game_mock.handle_request

        # Mock health check
        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_manager = TaskManager()

        # Load the frozen lake task definition for Fireworks
        task_def_path = Path("examples/frozen_lake/client/task_def.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        # Load and register the task
        task_def = task_manager._load_task_from_file(str(task_def_path))
        assert task_def is not None, "Failed to load task definition"

        # Override for traditional batch evaluation (not data-driven)
        task_def.dataset_path = (
            None  # Remove dataset path to use traditional evaluation
        )
        task_def.num_rollouts = 2

        task_id = task_manager.register_task(task_def)
        assert task_id == "frozen_lake_http_rollout"

        # Configure subprocess.Popen mock to prevent real process creation
        mock_process = Mock()
        mock_process.pid = 12345
        mock_process.poll.return_value = None
        mock_process.communicate.return_value = (b"", b"")
        mock_subprocess_popen.return_value = mock_process

        try:
            # Mock server process management
            with patch.object(
                task_manager, "_start_resource_server", return_value=12345
            ), patch.object(task_manager, "_wait_for_server_health", return_value=True):

                # Execute the task with batch evaluation
                results = await task_manager.execute_tasks(
                    task_ids=[task_id],
                    parallel=False,
                    max_concurrency=2,
                    num_rollouts_override=2,
                )

            # Validate results structure
            assert task_id in results
            result = results[task_id]

            # Should not be an error result
            assert not (
                isinstance(result, dict) and "error" in result
            ), f"Task failed: {result.get('error', 'Unknown error')}"

            # Should be aggregated results
            assert isinstance(result, dict)
            assert result.get(
                "aggregated", False
            ), "Results should be aggregated for batch evaluation"

            # Validate aggregated result structure
            required_keys = [
                "num_rollouts",
                "successful_rollouts",
                "success_rate",
                "avg_score",
                "min_score",
                "max_score",
            ]
            for key in required_keys:
                assert key in result, f"Missing key in aggregated results: {key}"

            # Validate result values
            assert result["num_rollouts"] == 2
            assert result["successful_rollouts"] >= 0
            assert result["successful_rollouts"] <= result["num_rollouts"]
            assert 0.0 <= result["success_rate"] <= 1.0
            assert isinstance(result["avg_score"], (int, float))
            assert isinstance(result["min_score"], (int, float))
            assert isinstance(result["max_score"], (int, float))
            assert result["min_score"] <= result["avg_score"] <= result["max_score"]

            # Should have individual results
            assert "individual_scores" in result
            assert "individual_results" in result
            assert len(result["individual_scores"]) == result["successful_rollouts"]
            assert len(result["individual_results"]) == result["successful_rollouts"]

            logging.info(f"Fireworks batch evaluation completed successfully: {result}")

        finally:
            await task_manager.cleanup()

    @pytest.mark.asyncio
    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("httpx.Client.post")
    @patch("requests.get")
    async def test_batch_evaluation_task_manager_openai(
        self, mock_requests_get, mock_httpx_post, mock_subprocess_popen, mock_openai
    ):
        """Test batch evaluation using TaskManager with OpenAI API."""
        # Mock OpenAI client and response
        mock_openai_client = AsyncMock()
        mock_openai.return_value = mock_openai_client

        mock_completion = AsyncMock()
        # Mock OpenAI completion response with proper tool call structure
        mock_tool_call = Mock()
        mock_tool_call.function.name = "step"
        mock_tool_call.function.arguments = '{"action": "down"}'
        mock_tool_call.id = "call_openai_123"

        mock_message = Mock()
        mock_message.content = None
        mock_message.role = "assistant"
        mock_message.tool_calls = [mock_tool_call]
        mock_message.model_dump = Mock(
            return_value={
                "role": "assistant",
                "content": None,
                "tool_calls": [
                    {
                        "id": "call_openai_123",
                        "type": "function",
                        "function": {"name": "step", "arguments": '{"action": "down"}'},
                    }
                ],
            }
        )

        mock_completion = AsyncMock()
        mock_completion.choices = [Mock(message=mock_message)]
        mock_completion.usage = Mock(total_tokens=15)
        mock_openai_client.chat.completions.create = AsyncMock(
            return_value=mock_completion
        )

        # Mock HTTP rollout server responses
        mock_httpx_response = Mock()
        mock_httpx_response.status_code = 200
        # Create responses for OpenAI test
        responses = [
            {"episode_id": "test_episode_openai"},
            {
                "position": [1, 0],
                "current_cell": "F",
                "done": False,
                "won": False,
                "message": "You moved down",
                "visual": " S  F  F  F\n[F] H  F  H\n F  F  F  H\n H  F  F  G",
            },
            {
                "position": [3, 3],
                "current_cell": "G",
                "done": True,
                "won": True,
                "message": "Victory!",
                "visual": " S  F  F  F\n F  H  F  H\n F  F  F  H\n H  F  F [G]",
            },
        ]
        response_iter = iter(responses * 10)
        mock_httpx_response.json.side_effect = lambda: next(
            response_iter, responses[-1]
        )
        mock_httpx_post.return_value = mock_httpx_response

        # Mock health check
        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_manager = TaskManager()

        # Load the frozen lake task definition for OpenAI
        task_def_path = Path("examples/frozen_lake/client/task_def_openai.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        # Load and register the task
        task_def = task_manager._load_task_from_file(str(task_def_path))
        assert task_def is not None, "Failed to load task definition"

        # Override num_rollouts to reduce test time
        task_def.num_rollouts = 2

        # Set OpenAI model temporarily
        original_model = os.environ.get("MODEL_AGENT")
        os.environ["MODEL_AGENT"] = "gpt-4o-mini"

        task_id = task_manager.register_task(task_def)
        assert task_id == "frozen_lake_http_rollout_openai"

        # Configure subprocess.Popen mock to prevent real process creation
        mock_process = Mock()
        mock_process.pid = 12346
        mock_process.poll.return_value = None
        mock_process.communicate.return_value = (b"", b"")
        mock_subprocess_popen.return_value = mock_process

        try:
            # Mock server process management
            with patch.object(
                task_manager, "_start_resource_server", return_value=12346
            ), patch.object(task_manager, "_wait_for_server_health", return_value=True):

                # Execute the task with batch evaluation
                results = await task_manager.execute_tasks(
                    task_ids=[task_id],
                    parallel=False,
                    max_concurrency=2,
                    num_rollouts_override=2,
                )

            # Validate results structure
            assert task_id in results
            result = results[task_id]

            # Should not be an error result
            assert not (
                isinstance(result, dict) and "error" in result
            ), f"Task failed: {result.get('error', 'Unknown error')}"

            # Should be aggregated results
            assert isinstance(result, dict)
            assert result.get(
                "aggregated", False
            ), "Results should be aggregated for batch evaluation"

            # Validate aggregated result structure
            required_keys = [
                "num_rollouts",
                "successful_rollouts",
                "success_rate",
                "avg_score",
                "min_score",
                "max_score",
            ]
            for key in required_keys:
                assert key in result, f"Missing key in aggregated results: {key}"

            # Validate result values
            assert result["num_rollouts"] == 2
            assert result["successful_rollouts"] >= 0
            assert result["successful_rollouts"] <= result["num_rollouts"]
            assert 0.0 <= result["success_rate"] <= 1.0
            assert isinstance(result["avg_score"], (int, float))
            assert isinstance(result["min_score"], (int, float))
            assert isinstance(result["max_score"], (int, float))
            assert result["min_score"] <= result["avg_score"] <= result["max_score"]

            # Should have individual results
            assert "individual_scores" in result
            assert "individual_results" in result
            assert len(result["individual_scores"]) == result["successful_rollouts"]
            assert len(result["individual_results"]) == result["successful_rollouts"]

            logging.info(f"OpenAI batch evaluation completed successfully: {result}")

        finally:
            # Restore original model
            if original_model:
                os.environ["MODEL_AGENT"] = original_model
            else:
                os.environ.pop("MODEL_AGENT", None)
            await task_manager.cleanup()

    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("aiohttp.ClientSession.post")
    @patch("httpx.Client.post")
    @patch("requests.get")
    def test_cli_batch_evaluation_fireworks(
        self,
        mock_requests_get,
        mock_httpx_post,
        mock_aiohttp_post,
        mock_subprocess_popen,
        mock_openai_constructor,
    ):
        """Test batch evaluation through CLI command with Fireworks."""
        # Mock OpenAI client in orchestrator
        mock_openai_client = AsyncMock()
        mock_openai_constructor.return_value = mock_openai_client

        mock_completion = AsyncMock()
        # Mock OpenAI completion response with proper tool call structure
        mock_tool_call = Mock()
        mock_tool_call.function.name = "step"
        mock_tool_call.function.arguments = '{"action": "up"}'
        mock_tool_call.id = "call_cli_fw"

        mock_message = Mock()
        mock_message.content = None
        mock_message.role = "assistant"
        mock_message.tool_calls = [mock_tool_call]
        mock_message.model_dump = Mock(
            return_value={
                "role": "assistant",
                "content": None,
                "tool_calls": [
                    {
                        "id": "call_cli_fw",
                        "type": "function",
                        "function": {"name": "step", "arguments": '{"action": "up"}'},
                    }
                ],
            }
        )

        mock_completion = AsyncMock()
        mock_completion.choices = [Mock(message=mock_message)]
        mock_completion.usage = Mock(total_tokens=8)
        mock_openai_client.chat.completions.create = AsyncMock(
            return_value=mock_completion
        )

        # Mock Fireworks API response (backup)
        mock_response = AsyncMock()
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value={
                "choices": [
                    {
                        "message": {"content": "up", "role": "assistant"},
                        "finish_reason": "stop",
                    }
                ],
                "usage": {"total_tokens": 8},
            }
        )
        mock_aiohttp_post.return_value.__aenter__.return_value = mock_response

        # Mock HTTP rollout responses
        mock_httpx_response = Mock()
        mock_httpx_response.status_code = 200

        responses = [
            {"episode_id": "test_cli_fw"},
            {"position": [0, 0], "done": False, "won": False, "message": "Test CLI FW"},
            {"position": [3, 3], "done": True, "won": True, "message": "CLI Win!"},
        ]
        response_iter = iter(responses * 5)
        mock_httpx_response.json.side_effect = lambda: next(
            response_iter, responses[-1]
        )
        mock_httpx_post.return_value = mock_httpx_response

        # Mock health check
        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_def_path = Path("examples/frozen_lake/client/task_def.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        # Create mock args for CLI command
        args = MockArgs(
            task_def=str(task_def_path),
            num_rollouts=2,
            parallel=False,
            max_concurrency=2,
        )

        # Configure subprocess.Popen mock to prevent real process creation
        mock_process = Mock()
        mock_process.pid = 12348
        mock_process.poll.return_value = None
        mock_process.communicate.return_value = (b"", b"")
        mock_subprocess_popen.return_value = mock_process

        # Execute CLI command with mocked subprocess for server management
        with patch("time.sleep"):
            exit_code = agent_eval_command(args)

        # Should complete successfully
        assert exit_code == 0, "CLI command should complete successfully"

    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("httpx.Client.post")
    @patch("requests.get")
    def test_cli_batch_evaluation_openai(
        self, mock_requests_get, mock_httpx_post, mock_subprocess_popen, mock_openai
    ):
        """Test batch evaluation through CLI command with OpenAI."""
        # Mock OpenAI client
        mock_openai_client = AsyncMock()
        mock_openai.return_value = mock_openai_client
        mock_completion = AsyncMock()
        # Mock OpenAI completion response with proper tool call structure
        mock_tool_call = Mock()
        mock_tool_call.function.name = "step"
        mock_tool_call.function.arguments = '{"action": "left"}'
        mock_tool_call.id = "call_cli_openai"

        mock_message = Mock()
        mock_message.content = None
        mock_message.role = "assistant"
        mock_message.tool_calls = [mock_tool_call]
        mock_message.model_dump = Mock(
            return_value={
                "role": "assistant",
                "content": None,
                "tool_calls": [
                    {
                        "id": "call_cli_openai",
                        "type": "function",
                        "function": {"name": "step", "arguments": '{"action": "left"}'},
                    }
                ],
            }
        )

        mock_completion = AsyncMock()
        mock_completion.choices = [Mock(message=mock_message)]
        mock_completion.usage = Mock(total_tokens=10)
        mock_openai_client.chat.completions.create = AsyncMock(
            return_value=mock_completion
        )

        # Mock HTTP rollout responses
        mock_httpx_response = Mock()
        mock_httpx_response.status_code = 200

        responses = [
            {"episode_id": "test_cli_openai"},
            {
                "position": [0, 0],
                "done": False,
                "won": False,
                "message": "Test CLI OpenAI",
            },
            {
                "position": [3, 3],
                "done": True,
                "won": True,
                "message": "CLI OpenAI Win!",
            },
        ]
        response_iter = iter(responses * 5)
        mock_httpx_response.json.side_effect = lambda: next(
            response_iter, responses[-1]
        )
        mock_httpx_post.return_value = mock_httpx_response

        # Mock health check
        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_def_path = Path("examples/frozen_lake/client/task_def_openai.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        # Set OpenAI model
        original_model = os.environ.get("MODEL_AGENT")
        os.environ["MODEL_AGENT"] = "gpt-4o-mini"

        try:
            # Create mock args for CLI command
            args = MockArgs(
                task_def=str(task_def_path),
                num_rollouts=2,
                parallel=False,
                max_concurrency=2,
            )

            # Configure subprocess.Popen mock to prevent real process creation
            mock_process = Mock()
            mock_process.pid = 12349
            mock_process.poll.return_value = None
            mock_process.communicate.return_value = (b"", b"")
            mock_subprocess_popen.return_value = mock_process

            # Execute CLI command with mocked subprocess for server management
            with patch("time.sleep"):
                exit_code = agent_eval_command(args)

            # Should complete successfully
            assert exit_code == 0, "CLI command should complete successfully"

        finally:
            # Restore original model
            if original_model:
                os.environ["MODEL_AGENT"] = original_model
            else:
                os.environ.pop("MODEL_AGENT", None)

    @pytest.mark.asyncio
    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("aiohttp.ClientSession.post")
    @patch("httpx.Client.post")
    @patch("requests.get")
    async def test_parallel_batch_evaluation(
        self,
        mock_requests_get,
        mock_httpx_post,
        mock_aiohttp_post,
        mock_subprocess_popen,
        mock_openai_constructor,
    ):
        """Test parallel execution of multiple rollouts."""
        # Mock OpenAI client in orchestrator
        mock_openai_client = AsyncMock()
        mock_openai_constructor.return_value = mock_openai_client

        # Mock OpenAI completion response with smart AI moves
        def create_tool_call_response(action, call_id):
            mock_tool_call = Mock()
            mock_tool_call.function.name = "step"
            mock_tool_call.function.arguments = f'{{"action": "{action}"}}'
            mock_tool_call.id = call_id

            mock_message = Mock()
            mock_message.content = None
            mock_message.role = "assistant"
            mock_message.tool_calls = [mock_tool_call]
            mock_message.model_dump = Mock(
                return_value={
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": call_id,
                            "type": "function",
                            "function": {
                                "name": "step",
                                "arguments": f'{{"action": "{action}"}}',
                            },
                        }
                    ],
                }
            )

            mock_completion = AsyncMock()
            mock_completion.choices = [Mock(message=mock_message)]
            mock_completion.usage = Mock(total_tokens=12)
            return mock_completion

        # Smart AI for parallel test
        winning_sequence = ["right", "right", "down", "down", "down", "right"]
        rollout_counter = [0]

        def smart_move_generator(**kwargs):
            messages = kwargs.get("messages", [])
            move_count = sum(
                1
                for msg in messages
                if msg.get("role") == "assistant" and msg.get("tool_calls")
            )

            rollout_counter[0] += 1
            if rollout_counter[0] <= 6:  # First rollout wins
                action = (
                    winning_sequence[move_count]
                    if move_count < len(winning_sequence)
                    else "right"
                )
            else:  # Second/third rollout makes mistake
                action = (
                    "right"
                    if move_count == 4
                    else (
                        winning_sequence[move_count]
                        if move_count < len(winning_sequence)
                        else "right"
                    )
                )

            return create_tool_call_response(action, f"call_{move_count}")

        mock_openai_client.chat.completions.create = AsyncMock(
            side_effect=smart_move_generator
        )

        # Mock Fireworks API response (backup)
        mock_response = AsyncMock()
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value={
                "choices": [
                    {
                        "message": {"content": "right", "role": "assistant"},
                        "finish_reason": "stop",
                    }
                ],
                "usage": {"total_tokens": 12},
            }
        )
        mock_aiohttp_post.return_value.__aenter__.return_value = mock_response

        # Use sophisticated game mock for parallel test
        game_mock = self._create_sophisticated_game_mock()
        mock_httpx_post.side_effect = game_mock.handle_request

        # Mock health check
        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_manager = TaskManager()

        # Load the frozen lake task definition
        task_def_path = Path("examples/frozen_lake/client/task_def.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        # Load and register the task
        task_def = task_manager._load_task_from_file(str(task_def_path))
        assert task_def is not None, "Failed to load task definition"

        # Test with more rollouts to verify parallelism
        task_def.num_rollouts = 3

        task_id = task_manager.register_task(task_def)

        # Configure subprocess.Popen mock to prevent real process creation
        mock_process = Mock()
        mock_process.pid = 12347
        mock_process.poll.return_value = None
        mock_process.communicate.return_value = (b"", b"")
        mock_subprocess_popen.return_value = mock_process

        try:
            # Mock server process management
            with patch.object(
                task_manager, "_start_resource_server", return_value=12347
            ), patch.object(task_manager, "_wait_for_server_health", return_value=True):

                # Execute with parallel enabled
                results = await task_manager.execute_tasks(
                    task_ids=[task_id],
                    parallel=True,
                    max_concurrency=2,
                    num_rollouts_override=3,
                )

            # Validate results
            assert task_id in results
            result = results[task_id]

            # Should be successful and aggregated
            assert not (isinstance(result, dict) and "error" in result)
            assert result.get("aggregated", False)
            assert result["num_rollouts"] == 3

            logging.info(f"Parallel batch evaluation completed: {result}")

        finally:
            await task_manager.cleanup()

    @pytest.mark.asyncio
    @patch("reward_kit.agent.orchestrator.AsyncOpenAI")
    @patch("subprocess.Popen")
    @patch("aiohttp.ClientSession.post")
    @patch("httpx.Client.post")
    @patch("requests.get")
    async def test_server_lifecycle_management(
        self,
        mock_requests_get,
        mock_httpx_post,
        mock_aiohttp_post,
        mock_subprocess_popen,
        mock_openai_constructor,
    ):
        """Test that resource servers are properly started and stopped."""
        # Mock OpenAI client in orchestrator
        mock_openai_client = AsyncMock()
        mock_openai_constructor.return_value = mock_openai_client

        # Mock OpenAI completion response with smart AI moves
        def create_tool_call_response(action, call_id):
            mock_tool_call = Mock()
            mock_tool_call.function.name = "step"
            mock_tool_call.function.arguments = f'{{"action": "{action}"}}'
            mock_tool_call.id = call_id

            mock_message = Mock()
            mock_message.content = None
            mock_message.role = "assistant"
            mock_message.tool_calls = [mock_tool_call]
            mock_message.model_dump = Mock(
                return_value={
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": call_id,
                            "type": "function",
                            "function": {
                                "name": "step",
                                "arguments": f'{{"action": "{action}"}}',
                            },
                        }
                    ],
                }
            )

            mock_completion = AsyncMock()
            mock_completion.choices = [Mock(message=mock_message)]
            mock_completion.usage = Mock(total_tokens=5)
            return mock_completion

        # Smart AI for lifecycle test
        winning_sequence = ["right", "right", "down", "down", "down", "right"]
        rollout_counter = [0]

        def smart_move_generator(**kwargs):
            messages = kwargs.get("messages", [])
            move_count = sum(
                1
                for msg in messages
                if msg.get("role") == "assistant" and msg.get("tool_calls")
            )

            rollout_counter[0] += 1
            if rollout_counter[0] <= 6:  # First rollout wins
                action = (
                    winning_sequence[move_count]
                    if move_count < len(winning_sequence)
                    else "right"
                )
            else:  # Second rollout makes mistake
                action = (
                    "right"
                    if move_count == 4
                    else (
                        winning_sequence[move_count]
                        if move_count < len(winning_sequence)
                        else "right"
                    )
                )

            return create_tool_call_response(action, f"call_{move_count}")

        mock_openai_client.chat.completions.create = AsyncMock(
            side_effect=smart_move_generator
        )

        # Mock API responses (backup)
        mock_response = AsyncMock()
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value={
                "choices": [{"message": {"content": "down", "role": "assistant"}}],
                "usage": {"total_tokens": 5},
            }
        )
        mock_aiohttp_post.return_value.__aenter__.return_value = mock_response

        # Use sophisticated game mock for lifecycle test
        game_mock = self._create_sophisticated_game_mock()
        mock_httpx_post.side_effect = game_mock.handle_request

        mock_health_response = Mock()
        mock_health_response.status_code = 200
        mock_requests_get.return_value = mock_health_response

        task_manager = TaskManager()

        # Load task definition
        task_def_path = Path("examples/frozen_lake/client/task_def.yaml")
        if not task_def_path.exists():
            pytest.skip(f"Task definition not found: {task_def_path}")

        task_def = task_manager._load_task_from_file(str(task_def_path))
        task_def.num_rollouts = 2
        task_id = task_manager.register_task(task_def)

        # Check that no servers are running initially
        assert len(task_manager.server_processes) == 0
        assert len(task_manager.server_ports) == 0

        # Configure subprocess.Popen mock to prevent real process creation
        mock_process = Mock()
        mock_process.pid = 12348
        mock_process.poll.return_value = None
        mock_process.communicate.return_value = (b"", b"")
        mock_subprocess_popen.return_value = mock_process

        try:
            # Mock server process management
            with patch.object(
                task_manager, "_start_resource_server", return_value=12348
            ), patch.object(task_manager, "_wait_for_server_health", return_value=True):

                # Execute task
                results = await task_manager.execute_tasks(
                    task_ids=[task_id], num_rollouts_override=2
                )

            # Task should complete successfully
            assert task_id in results
            result = results[task_id]
            assert not (isinstance(result, dict) and "error" in result)

        finally:
            await task_manager.cleanup()

            # Check that all servers are cleaned up
            assert len(task_manager.server_processes) == 0
            assert len(task_manager.server_ports) == 0


@pytest.mark.integration
class TestBatchEvaluationErrorHandling:
    """Test error handling in batch evaluation scenarios."""

    @pytest.mark.asyncio
    async def test_invalid_task_definition(self):
        """Test handling of invalid task definitions."""
        task_manager = TaskManager()

        # Create a temporary invalid task definition
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(
                """
name: "invalid_task"
description: "This task has invalid configuration"
resource_type: "nonexistent_resource"
"""
            )
            invalid_task_path = f.name

        try:
            # Attempt to load invalid task
            task_def = task_manager._load_task_from_file(invalid_task_path)

            # Should either fail to load or fail during execution
            if task_def is not None:
                task_id = task_manager.register_task(task_def)
                results = await task_manager.execute_tasks([task_id])

                # Should result in error
                assert task_id in results
                result = results[task_id]
                assert isinstance(result, dict) and "error" in result

        finally:
            # Clean up temporary file
            Path(invalid_task_path).unlink(missing_ok=True)
            await task_manager.cleanup()

    @pytest.mark.asyncio
    async def test_missing_api_key_handling(self):
        """Test graceful handling when API keys are missing."""
        # Temporarily remove API keys
        original_fw_key = os.environ.get("FIREWORKS_API_KEY")
        original_openai_key = os.environ.get("OPENAI_API_KEY")

        os.environ.pop("FIREWORKS_API_KEY", None)
        os.environ.pop("OPENAI_API_KEY", None)

        task_manager = TaskManager()

        try:
            task_def_path = Path("examples/frozen_lake/client/task_def.yaml")
            if not task_def_path.exists():
                pytest.skip(f"Task definition not found: {task_def_path}")

            task_def = task_manager._load_task_from_file(str(task_def_path))
            if task_def:
                task_def.num_rollouts = 1  # Reduce rollouts for faster failure
                task_id = task_manager.register_task(task_def)

                results = await task_manager.execute_tasks([task_id])

                # Should handle missing API key gracefully
                assert task_id in results
                result = results[task_id]
                # Result could be error or have low success rate due to API failures
                if isinstance(result, dict) and "error" in result:
                    # Direct error is acceptable
                    pass
                elif isinstance(result, dict) and result.get("aggregated", False):
                    # Batch result with low success rate is also acceptable
                    assert result["success_rate"] <= 1.0

        finally:
            # Restore API keys
            if original_fw_key:
                os.environ["FIREWORKS_API_KEY"] = original_fw_key
            if original_openai_key:
                os.environ["OPENAI_API_KEY"] = original_openai_key
            await task_manager.cleanup()



import asyncio
import json
import logging
from contextlib import AsyncExitStack
from typing import Any, Dict, List, Optional

import aiohttp  # Still needed for type hints if we expose the session, but primary interaction changes
import mcp.types  # Reverted to mcp.types; Explicit import for clarity
from mcp import types as mcp_types  # Reverted to mcp.types; Explicit import for clarity
from mcp.client.session import DEFAULT_CLIENT_INFO, ClientSession
from mcp.client.streamable_http import streamablehttp_client
from omegaconf import DictConfig

logger = logging.getLogger(__name__)


class IntermediaryMCPClient:
    """
    Client for interacting with the RewardKitIntermediaryServer using mcp.client components.
    """

    def __init__(self, intermediary_server_url: str):
        if not intermediary_server_url:
            raise ValueError("intermediary_server_url must be provided.")
        self.server_url = intermediary_server_url.rstrip(
            "/"
        )  # Should be like http://localhost:8001/mcp

        self._exit_stack: Optional[AsyncExitStack] = None
        self._mcp_session: Optional[ClientSession] = None

    async def connect(self):
        """Establishes connection and MCP session."""
        if self._mcp_session and not self._mcp_session.is_closed:
            logger.debug("Already connected.")
            return

        self._exit_stack = AsyncExitStack()
        try:
            logger.debug(
                f"Attempting to connect to Intermediary MCP server at {self.server_url}"
            )
            read_stream, write_stream, http_session_info = (
                await self._exit_stack.enter_async_context(
                    streamablehttp_client(self.server_url)
                )
            )
            # http_session_info might contain the underlying aiohttp session if needed, and mcp_session_id
            # logger.debug(f"Streamable HTTP transport established. HTTP session info: {http_session_info}")

            self._mcp_session = await self._exit_stack.enter_async_context(
                ClientSession(
                    read_stream, write_stream, client_info=DEFAULT_CLIENT_INFO
                )
            )
            await self._mcp_session.initialize()
            logger.info(
                f"IntermediaryMCPClient connected and MCP session initialized with {self.server_url}"
            )
        except Exception as e:
            if self._exit_stack:  # pragma: no cover
                await self._exit_stack.aclose()
                self._exit_stack = None
            self._mcp_session = None
            logger.error(
                f"Failed to connect or initialize MCP session with {self.server_url}: {e}",
                exc_info=True,
            )
            raise

    async def close(self):
        """Closes the MCP session and underlying transport."""
        if self._exit_stack:
            logger.debug(
                f"Closing IntermediaryMCPClient connection to {self.server_url}"
            )
            await self._exit_stack.aclose()
            self._exit_stack = None
            self._mcp_session = None
            logger.info(
                f"IntermediaryMCPClient connection to {self.server_url} closed."
            )

    async def _ensure_connected(self):
        # ClientSession doesn't have a public is_closed.
        # We rely on _mcp_session being None or connect() re-establishing.
        # The AsyncExitStack handles actual closure of resources.
        if not self._mcp_session:
            logger.debug("Session not established, attempting to connect...")
            await self.connect()

        # After attempting to connect, if _mcp_session is still None, it means connection failed.
        if not self._mcp_session:
            raise RuntimeError("Failed to establish or re-establish MCP session.")

    async def _call_intermediary_tool(
        self, tool_name: str, tool_args_payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Helper to make a raw tool call to the intermediary server and parse the result.
        The tool_args_payload is the "arguments" field for the intermediary's tool.
        """
        await self._ensure_connected()
        if not self._mcp_session:  # For type checker
            raise RuntimeError("MCP session not available after ensure_connected.")

        logger.debug(
            f"Calling intermediary tool '{tool_name}' with payload: {tool_args_payload}"
        )

        mcp_response: mcp_types.CallToolResult = await self._mcp_session.call_tool(
            tool_name, tool_args_payload
        )

        logger.debug(
            f"Raw MCP response from intermediary for '{tool_name}': {mcp_response}"
        )

        if (
            mcp_response.isError
            or not mcp_response.content
            or not hasattr(mcp_response.content[0], "text")
        ):
            error_message = f"Tool call '{tool_name}' to intermediary failed."
            if (
                mcp_response.isError
                and mcp_response.content
                and hasattr(mcp_response.content[0], "text")
            ):
                error_message += f" Details: {mcp_response.content[0].text}"
            elif mcp_response.isError:
                error_message += " No detailed error message in content."
            logger.error(error_message)
            try:
                if mcp_response.content and hasattr(mcp_response.content[0], "text"):
                    parsed_error = json.loads(mcp_response.content[0].text)
                    if isinstance(parsed_error, dict) and "error" in parsed_error:
                        raise RuntimeError(
                            f"{error_message} Nested error: {parsed_error['error']}"
                        )
            except (json.JSONDecodeError, TypeError):
                pass
            raise RuntimeError(error_message)

        try:
            parsed_result = json.loads(mcp_response.content[0].text)
            logger.debug(
                f"Parsed JSON result from intermediary for '{tool_name}': {parsed_result}"
            )
            return parsed_result
        except json.JSONDecodeError as e:
            logger.error(
                f"Failed to parse JSON from intermediary's tool '{tool_name}' response content: {mcp_response.content[0].text}. Error: {e}"
            )
            raise RuntimeError(
                f"Failed to parse JSON response from intermediary tool '{tool_name}'."
            )

    async def initialize_session(
        self, backend_requests: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """
        Initializes a session with the IntermediaryServer, requesting backend instances.
        """
        payload_for_intermediary_tool = {"args": {"backends": backend_requests}}
        return await self._call_intermediary_tool(
            tool_name="initialize_session",
            tool_args_payload=payload_for_intermediary_tool,
        )

    async def call_backend_tool(
        self,
        rk_session_id: str,
        instance_id: str,
        backend_name_ref: str,
        tool_name: str,
        tool_args: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Calls a tool on a specific backend instance managed by the IntermediaryServer.
        """
        payload_for_intermediary_tool = {
            "args": {
                "rk_session_id": rk_session_id,
                "instance_id": instance_id,
                "backend_name_ref": backend_name_ref,
                "tool_name": tool_name,
                "tool_args": tool_args,
            }
        }
        return await self._call_intermediary_tool(
            tool_name="call_backend_tool",
            tool_args_payload=payload_for_intermediary_tool,
        )

    async def list_backend_tools(
        self, rk_session_id: str, instance_id: str, backend_name_ref: str
    ) -> mcp_types.ListToolsResult:
        """
        Lists tools available on a specific backend instance via the IntermediaryServer.
        """
        payload_for_intermediary_tool = {
            "args": {
                "rk_session_id": rk_session_id,
                "instance_id": instance_id,
                "backend_name_ref": backend_name_ref,
            }
        }
        # _call_intermediary_tool returns a Dict[str, Any] which is the parsed JSON
        # from the intermediary's response. This dict should be the model_dump of ListToolsResult.
        raw_result_dict = await self._call_intermediary_tool(
            tool_name="list_backend_tools",
            tool_args_payload=payload_for_intermediary_tool,
        )
        # Parse the dictionary back into the Pydantic model
        return mcp_types.ListToolsResult(**raw_result_dict)

    async def cleanup_session(self, rk_session_id: str) -> Dict[str, Any]:
        """
        Cleans up a session on the IntermediaryServer.
        """
        payload_for_intermediary_tool = {"args": {"rk_session_id": rk_session_id}}
        result = await self._call_intermediary_tool(
            tool_name="cleanup_session", tool_args_payload=payload_for_intermediary_tool
        )
        return result

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



"""Deprecated adapter wrappers for Braintrust.

This module forwards imports to :mod:`reward_kit.integrations.braintrust`.
"""

from ..integrations.braintrust import reward_fn_to_scorer, scorer_to_reward_fn

__all__ = ["scorer_to_reward_fn", "reward_fn_to_scorer"]



import logging
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Tuple

logger = logging.getLogger(__name__)


def _run_gcloud_command(
    command: List[str], dry_run: bool = False
) -> Tuple[bool, str, str]:
    """
    Helper to run a gcloud command.
    In a real scenario, this would interact with subprocess.
    Returns: (success_status, stdout, stderr)
    """
    command_str_for_print = " ".join(["gcloud"] + command)
    logger.info(f"Executing: {command_str_for_print}")
    if dry_run:
        logger.info(f"Dry run mode. Command not executed: {command_str_for_print}")
        return True, f"Dry run: {command_str_for_print}", ""

    try:
        process = subprocess.run(
            ["gcloud"] + command,
            capture_output=True,
            text=True,
            check=False,  # Handle non-zero exit codes manually
        )
        if process.returncode == 0:
            if (
                process.stderr
            ):  # gcloud sometimes prints informational messages to stderr on success
                logger.info(f"Command successful with stderr output:\n{process.stderr}")
            return True, process.stdout.strip(), process.stderr.strip()
        else:
            logger.error(f"Command failed. Return code: {process.returncode}")
            logger.error(f"Stdout:\n{process.stdout}")
            logger.error(f"Stderr:\n{process.stderr}")
            return False, process.stdout.strip(), process.stderr.strip()
    except FileNotFoundError:
        logger.error("gcloud command not found. Is it installed and in PATH?")
        return False, "", "gcloud command not found."
    except Exception as e:
        logger.error(f"An unexpected error occurred while running gcloud command: {e}")
        return False, "", str(e)


def build_and_push_docker_image(
    image_name_tag: str,  # e.g., gcr.io/my-project/my-reward-func:latest
    dockerfile_content: str,
    build_context_dir: str,  # Directory where Dockerfile and user code are (usually CWD)
    gcp_project_id: Optional[
        str
    ] = None,  # Required if using gcloud builds submit without local Docker
    dry_run: bool = False,
) -> bool:
    """
    Builds a Docker image using the provided Dockerfile content and pushes it to a registry (e.g., GCR, Artifact Registry).
    Can use local Docker or 'gcloud builds submit'.

    Args:
        image_name_tag: Full name and tag for the image (e.g., "gcr.io/project-id/image-name:tag").
        dockerfile_content: String content of the Dockerfile.
        build_context_dir: The build context directory for Docker.
        gcp_project_id: GCP Project ID, used for 'gcloud builds submit'.
        dry_run: If True, prints commands instead of executing them.

    Returns:
        True if successful, False otherwise.
    """
    logger.info(
        f"Attempting to build and push Docker image using Google Cloud Build: {image_name_tag}"
    )

    if not gcp_project_id:
        logger.error("GCP Project ID is required for Google Cloud Build.")
        return False

    # Create Dockerfile in the build_context_dir. It must be named "Dockerfile".
    dockerfile_path_in_context = Path(build_context_dir) / "Dockerfile"

    try:
        with open(dockerfile_path_in_context, "w") as f:
            f.write(dockerfile_content)
        logger.info(f"Dockerfile created at: {dockerfile_path_in_context}")

        # Command for gcloud builds submit
        # The build_context_dir (e.g., ".") is where gcloud looks for the Dockerfile and other source files.
        build_cmd_gcloud = [
            "builds",
            "submit",
            build_context_dir,  # Source code to upload (can be "." for CWD)
            "--tag",
            image_name_tag,
            "--project",
            gcp_project_id,
        ]

        success, stdout, stderr = _run_gcloud_command(build_cmd_gcloud, dry_run=dry_run)

        if not success:
            logger.error(
                f"Google Cloud Build failed. Stdout: {stdout}, Stderr: {stderr}"
            )
            return False

    except Exception as e:
        logger.error(
            f"An error occurred during Dockerfile creation or gcloud command preparation: {e}"
        )
        return False
    finally:
        if dockerfile_path_in_context.exists():
            os.remove(dockerfile_path_in_context)
            logger.info(f"Temporary Dockerfile {dockerfile_path_in_context} removed.")

    if success:
        logger.info(f"Successfully built and pushed image {image_name_tag}")
    else:
        logger.error(f"Failed to build and push image {image_name_tag}")
    return success


def deploy_to_cloud_run(
    service_name: str,
    image_name_tag: str,
    gcp_project_id: str,
    gcp_region: str,
    allow_unauthenticated: bool = True,  # For --auth api-key, the service itself is open, auth is app-level
    env_vars: Optional[Dict[str, str]] = None,
    secrets_to_mount: Optional[Dict[str, str]] = None,
    service_port: int = 8080,
    dry_run: bool = False,
) -> Optional[str]:
    """
    Deploys a container image to Google Cloud Run.

    Args:
        service_name: Name for the Cloud Run service.
        image_name_tag: Full name of the Docker image to deploy (e.g., "gcr.io/project/image:tag").
        gcp_project_id: GCP Project ID.
        gcp_region: GCP Region for the service.
        allow_unauthenticated: Whether to allow unauthenticated invocations (publicly accessible).
        env_vars: Environment variables to set for the service.
        secrets_to_mount: Secrets from GCP Secret Manager to mount as environment variables.
        service_port: Port the container exposes.
        dry_run: If True, prints commands instead of executing them.

    Returns:
        The URL of the deployed service if successful, else None.
    """

    if not gcp_project_id:
        logger.error("GCP Project ID is required for deploying to Cloud Run.")
        return None
    if not gcp_region:
        logger.error("GCP Region is required for deploying to Cloud Run.")
        return None

    try:
        logger.info(
            f"Deploying image {image_name_tag} to Cloud Run service {service_name} in {gcp_region} (Project: {gcp_project_id})"
        )

        deploy_cmd_list = [
            "run",
            "deploy",
            service_name,
            "--image",
            image_name_tag,
            "--region",
            gcp_region,
            "--project",
            gcp_project_id,
            "--port",
            str(service_port),
            # "--platform", "managed",
        ]

        if allow_unauthenticated:
            deploy_cmd_list.append("--allow-unauthenticated")
        else:
            # For IAM based auth, would be --no-allow-unauthenticated and then set IAM policy
            deploy_cmd_list.append("--no-allow-unauthenticated")
            logger.info(
                "Note: --no-allow-unauthenticated set. Further IAM configuration might be needed."
            )

        if env_vars:
            env_vars_str = ",".join([f"{k}={v}" for k, v in env_vars.items()])
            deploy_cmd_list.extend(["--set-env-vars", env_vars_str])

        if secrets_to_mount:
            # Format: ENV_VAR_NAME=secret_name:version,...
            # Secret name here is just the short ID, not the full path.
            # gcloud will resolve it within the project.
            # Example: MY_API_KEY=my-api-key-secret:latest
            secrets_str_list = []
            for env_var_name, secret_manager_full_id in secrets_to_mount.items():
                # Parse projects/PROJECT_ID/secrets/SECRET_ID/versions/VERSION
                parts = secret_manager_full_id.split("/")
                if (
                    len(parts) == 6
                    and parts[0] == "projects"
                    and parts[2] == "secrets"
                    and parts[4] == "versions"
                ):
                    secret_id = parts[3]
                    secret_version = parts[5]
                    secrets_str_list.append(
                        f"{env_var_name}={secret_id}:{secret_version}"
                    )
                else:
                    logger.warning(
                        f"Invalid secret manager full ID format: {secret_manager_full_id}. Skipping secret mount for {env_var_name}."
                    )

            if secrets_str_list:
                deploy_cmd_list.extend(["--update-secrets", ",".join(secrets_str_list)])

        success, stdout, stderr = _run_gcloud_command(deploy_cmd_list, dry_run=dry_run)

        if success:
            if dry_run:
                service_url_placeholder = f"https://{service_name}-mock-url.a.run.app"
                logger.info(
                    f"Successfully deployed service {service_name} (dry run). URL (placeholder): {service_url_placeholder}"
                )
                return service_url_placeholder

            # Get the service URL after successful deployment
            get_url_cmd = [
                "run",
                "services",
                "describe",
                service_name,
                "--region",
                gcp_region,
                "--project",
                gcp_project_id,
                "--format",
                "value(status.url)",
            ]
            url_success, url_stdout, url_stderr = _run_gcloud_command(
                get_url_cmd, dry_run=False
            )  # Always try to get URL if deploy was not dry_run

            if url_success and url_stdout:
                service_url = url_stdout.strip()
                if not service_url.startswith("https://"):
                    logger.error(
                        f"Service URL is not valid (must be HTTPS): {service_url}"
                    )
                    return None
                logger.info(
                    f"Successfully deployed service {service_name}. URL: {service_url}"
                )
                return service_url
            else:
                logger.error(
                    f"Deployed service {service_name}, but failed to retrieve its URL. Stderr: {url_stderr}"
                )
                return None  # Consider deployment failed if URL cannot be retrieved
        else:
            logger.error(f"Failed to deploy service {service_name}. Stderr: {stderr}")
            return None
    except Exception as e:
        logger.error(
            f"An error occurred during Cloud Run deployment for service {service_name}: {e}"
        )
        return None


def ensure_artifact_registry_repo_exists(
    project_id: str, region: str, repo_name: str, dry_run: bool = False
) -> bool:
    """
    Checks if an Artifact Registry repository exists, and creates it if it doesn't.
    """
    logger.info(
        f"Ensuring Artifact Registry repository '{repo_name}' exists in project '{project_id}', region '{region}'."
    )

    try:
        describe_cmd = [
            "artifacts",
            "repositories",
            "describe",
            repo_name,
            "--project",
            project_id,
            "--location",
            region,
        ]

        # Don't use dry_run for describe, as we need to know if it exists
        success, stdout, stderr = _run_gcloud_command(describe_cmd, dry_run=False)

        if success:
            logger.info(f"Artifact Registry repository '{repo_name}' already exists.")
            return True

        # If describe failed, check if it's because the repo was not found
        # gcloud typically returns non-zero exit code and an error message to stderr for "not found"
        if (
            "NOT_FOUND" in stderr.upper() or "failed to find" in stderr.lower()
        ):  # Heuristic check
            logger.info(
                f"Artifact Registry repository '{repo_name}' not found. Attempting to create it."
            )
            create_cmd = [
                "artifacts",
                "repositories",
                "create",
                repo_name,
                "--project",
                project_id,
                "--repository-format",
                "docker",
                "--location",
                region,
                "--description",
                "Repository for reward-kit evaluators (auto-created by reward-kit CLI)",
            ]
            create_success, create_stdout, create_stderr = _run_gcloud_command(
                create_cmd, dry_run=dry_run
            )
            if create_success:
                logger.info(
                    f"Successfully created Artifact Registry repository '{repo_name}'."
                )
                return True
            else:
                logger.error(
                    f"Failed to create Artifact Registry repository '{repo_name}'. Stderr: {create_stderr}"
                )
                return False
        else:
            # Describe failed for a reason other than "not found"
            logger.error(
                f"Error describing Artifact Registry repository '{repo_name}'. Stderr: {stderr}"
            )
            return False
    except Exception as e:
        logger.error(
            f"An unexpected error occurred while ensuring Artifact Registry repository '{repo_name}': {e}"
        )
        return False


def ensure_gcp_secret(
    project_id: str,
    secret_id: str,
    secret_value: str,
    region: Optional[
        str
    ] = None,  # For replication policy if needed, or if secrets are regional
    labels: Optional[Dict[str, str]] = None,
    dry_run: bool = False,
) -> Optional[str]:
    """
    Ensures a secret exists in GCP Secret Manager and adds the given value as a new version.
    Returns the full resource name of the new secret version if successful, else None.
    e.g., projects/PROJECT_ID/secrets/SECRET_ID/versions/VERSION
    """
    if not project_id:
        logger.error("GCP Project ID is required to manage secrets.")
        return None
    if not secret_id:
        logger.error("Secret ID is required to manage secrets.")
        return None
    if secret_value is None:
        logger.error("Secret value is required to create or update a secret.")
        return None

    logger.info(f"Ensuring secret '{secret_id}' in project '{project_id}'.")
    describe_cmd = ["secrets", "describe", secret_id, "--project", project_id]
    secret_exists, _, describe_stderr = _run_gcloud_command(describe_cmd, dry_run=False)

    if not secret_exists:
        if (
            "NOT_FOUND" in describe_stderr.upper()
            or "failed to find" in describe_stderr.lower()
        ):
            logger.info(f"Secret '{secret_id}' not found. Attempting to create it.")
            create_cmd_list = [
                "secrets",
                "create",
                secret_id,
                "--project",
                project_id,
            ]
            # Replication policy: automatic is default (global).
            # If a region is provided, could set --replication-policy user-managed --locations <region>
            # For simplicity, using automatic for now.
            # TODO: Consider if region-specific replication is needed.
            # if region:
            #     create_cmd_list.extend(["--replication-policy", "user-managed", "--locations", region])
            # else:
            create_cmd_list.extend(["--replication-policy", "automatic"])

            if labels:
                labels_str = ",".join([f"{k}={v}" for k, v in labels.items()])
                create_cmd_list.extend(["--labels", labels_str])

            create_success, _, create_stderr = _run_gcloud_command(
                create_cmd_list, dry_run=dry_run
            )
            if not create_success:
                logger.error(
                    f"Failed to create secret '{secret_id}'. Stderr: {create_stderr}"
                )
                return None
            logger.info(f"Successfully created secret '{secret_id}'.")
            secret_exists = True  # Now it exists
        else:
            # Describe failed for another reason
            logger.error(
                f"Error describing secret '{secret_id}'. Stderr: {describe_stderr}"
            )
            return None

    # Add a new version to the secret
    # Create a temporary file for the secret value
    try:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_secret_file:
            tmp_secret_file.write(secret_value)
            tmp_secret_file_path = tmp_secret_file.name

        add_version_cmd = [
            "secrets",
            "versions",
            "add",
            secret_id,
            "--project",
            project_id,
            "--data-file",
            tmp_secret_file_path,
        ]
        version_success, version_stdout, version_stderr = _run_gcloud_command(
            add_version_cmd, dry_run=dry_run
        )

        if tmp_secret_file_path and os.path.exists(tmp_secret_file_path):
            os.remove(tmp_secret_file_path)

        if not version_success:
            logger.error(
                f"Failed to add version to secret '{secret_id}'. Stderr: {version_stderr}"
            )
            return None

        # The stdout of 'versions add' usually contains the version name, but it's safer to describe.
        # Let's parse the version from the output if available, or describe to get the latest.
        # For simplicity, if dry_run, we can't get a real version.
        if dry_run:
            logger.info(
                f"Successfully added version to secret '{secret_id}' (dry run)."
            )
            return f"projects/{project_id}/secrets/{secret_id}/versions/latest-dry-run"

        # Get the full name of the newly added version
        # 'gcloud secrets versions describe latest --secret=SECRET_ID --format="value(name)"' gets the name
        describe_version_cmd = [
            "secrets",
            "versions",
            "describe",
            "latest",
            "--secret",
            secret_id,
            "--project",
            project_id,
            "--format",
            "value(name)",
        ]
        desc_ver_success, desc_ver_stdout, desc_ver_stderr = _run_gcloud_command(
            describe_version_cmd, dry_run=False
        )
        if desc_ver_success and desc_ver_stdout:
            secret_version_name = desc_ver_stdout.strip()
            logger.info(
                f"Successfully added version to secret '{secret_id}'. Version name: {secret_version_name}"
            )
            return secret_version_name
        else:
            logger.error(
                f"Added version to secret '{secret_id}', but failed to retrieve new version name. Stderr: {desc_ver_stderr}"
            )
            return None

    except Exception as e:
        logger.error(f"An error occurred while adding secret version: {e}")
        if "tmp_secret_file_path" in locals() and os.path.exists(tmp_secret_file_path):
            os.remove(tmp_secret_file_path)
        return None


if __name__ == "__main__":
    # Basic setup for logger to see output when run directly
    logging.basicConfig(level=logging.INFO, format="%(levelname)s:%(name)s:%(message)s")
    logger.info("--- GCP Tools Module (Placeholder Examples) ---")

    # Note: Dockerfile content would come from packaging.py
    dummy_dockerfile = 'FROM python:3.10-slim\nCMD ["echo", "hello"]'
    img_name = "gcr.io/my-test-project/my-test-reward-eval:latest"  # Old GCR name, update for AR
    # Example AR image name: us-central1-docker.pkg.dev/my-test-project/my-ar-repo/my-test-reward-eval:latest
    ar_img_name = "us-central1-docker.pkg.dev/my-test-project/reward-kit-images/my-test-reward-eval:latest"

    print(f"\n1. Simulating build and push for {ar_img_name} (dry_run=True)")
    build_and_push_docker_image(
        image_name_tag=ar_img_name,
        dockerfile_content=dummy_dockerfile,
        build_context_dir=".",  # Assumes CWD is build context
        gcp_project_id="my-test-project",
        dry_run=True,
    )

    print(f"\n2. Simulating deploy to Cloud Run (dry_run=True)")
    deploy_to_cloud_run(
        service_name="my-reward-service",
        image_name_tag=ar_img_name,  # Use AR image name
        gcp_project_id="my-test-project",
        gcp_region="us-central1",
        allow_unauthenticated=True,
        env_vars={"MY_ENV_VAR": "my_value"},
        secrets_to_mount={
            "API_KEY_SECRET": "projects/my-test-project/secrets/my-api-key/versions/latest"
        },
        dry_run=True,
    )

    print(f"\n3. Simulating ensure_artifact_registry_repo_exists (dry_run=True)")
    ensure_artifact_registry_repo_exists(
        project_id="my-test-project",
        region="us-central1",
        repo_name="reward-kit-evaluators",
        dry_run=True,
    )

    print(f"\n4. Simulating ensure_gcp_secret (dry_run=True)")
    ensure_gcp_secret(
        project_id="my-test-project",
        secret_id="my-test-api-key-secret",
        secret_value="supersecretvalue123",
        labels={"managed-by": "reward-kit-test"},
        dry_run=True,
    )
    print(
        "\nNote: These are placeholder executions. Real implementation requires gcloud CLI and Docker."
    )



"""
Out-of-the-box reward functions for common use cases.
"""

# Import specific reward functions
from . import (
    accuracy,
    accuracy_length,
    bfcl_reward,
    code_execution,
    cpp_code,
    deepcoder_reward,
    format,
    function_calling,
    json_schema,
    language_consistency,
    lean_prover,
    length,
    list_comparison_math_reward,
    math,
    multiple_choice_math_reward,
    reasoning_steps,
    repetition,
    tag_count,
)
from .accuracy_length import cosine_scaled_accuracy_length_reward

# Import function separately to avoid name conflict with the module
from .bfcl_reward import bfcl_reward as bfcl_reward_function

# Directly import specific reward functions for easy access
from .code_execution import fractional_code_reward
from .cpp_code import binary_cpp_code_reward, ioi_cpp_code_reward
from .deepcoder_reward import deepcoder_code_reward

# To make individual functions directly importable from reward_kit.rewards
# e.g., from reward_kit.rewards import composite_function_call_reward
from .function_calling import (
    composite_function_call_reward,
    exact_tool_match_reward,
    llm_judge_reward,
    schema_jaccard_reward,
)
from .lean_prover import (
    deepseek_huggingface_prover_benchmark,
    deepseek_prover_v2_reward,
    lean_prover_reward,
)

# Import these with aliases to avoid name conflicts
from .list_comparison_math_reward import (
    list_comparison_math_reward as list_comparison_math_reward_function,
)
from .multiple_choice_math_reward import (
    multiple_choice_math_reward as multiple_choice_math_reward_function,
)

__all__ = [
    # Modules
    "function_calling",
    "json_schema",
    "math",
    "code_execution",
    "format",
    "tag_count",
    "accuracy",
    "language_consistency",
    "reasoning_steps",
    "length",
    "repetition",
    "cpp_code",
    "accuracy_length",
    "lean_prover",
    "deepcoder_reward",
    "multiple_choice_math_reward",
    "list_comparison_math_reward",
    "bfcl_reward",
    # Specific functions for direct import
    "fractional_code_reward",
    "deepcoder_code_reward",
    "multiple_choice_math_reward_function",
    "list_comparison_math_reward_function",
    "ioi_cpp_code_reward",
    "binary_cpp_code_reward",
    "cosine_scaled_accuracy_length_reward",
    "lean_prover_reward",
    "deepseek_prover_v2_reward",
    "deepseek_huggingface_prover_benchmark",
    "bfcl_reward_function",
    "composite_function_call_reward",
    "exact_tool_match_reward",
    "llm_judge_reward",
    "schema_jaccard_reward",
]



"""
C/C++ code execution reward functions for evaluating C/C++ code correctness.

This module provides functions to evaluate the correctness of C/C++ code by:
1. Extracting code blocks from messages
2. Executing the code using the Piston execution engine
3. Comparing the output with expected results or running against test cases
"""

import asyncio
import json
import os
import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import aiohttp

from ..models import EvaluateResult, Message, MetricResult
from ..reward_function import reward_function


@dataclass
class TestResult:
    """
    Represents the result of a single test case execution.
    """

    test_name: str
    score: float = 0.0
    status: str = "SKIPPED"
    feedback: str = ""
    actual_output: str = ""
    expected_output: str = ""


class PistonError(Exception):
    """Exception raised for errors from the Piston API."""

    pass


class PistonClient:
    """
    A client that communicates with Piston API endpoints for code execution.

    Piston is a general purpose code execution engine:
    https://github.com/engineer-man/piston
    """

    def __init__(
        self,
        base_endpoint: str = "https://emkc.org/api/v2/piston",
        session: Optional[aiohttp.ClientSession] = None,
        timeout: int = 30,
    ):
        self.base_endpoint = base_endpoint
        self._session = session
        self.timeout = timeout

    @property
    def session(self):
        if self._session is None:
            self._session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(sock_read=self.timeout),
                connector=aiohttp.TCPConnector(
                    limit=10,
                    ttl_dns_cache=300,
                    keepalive_timeout=30,
                ),
            )
        return self._session

    async def close(self):
        """Close the session."""
        if self._session:
            await self._session.close()
            self._session = None

    async def get_runtimes(self) -> List[Dict[str, Any]]:
        """Get list of supported runtimes."""
        async with self.session.get(f"{self.base_endpoint}/runtimes") as response:
            if response.status != 200:
                raise PistonError(f"Error getting runtimes: {response.status}")
            return await response.json()

    async def execute(
        self,
        language: str,
        version: str,
        files: List[Dict[str, str]],
        stdin: str = "",
        args: List[str] = [],
        compile_timeout: int = 10000,
        run_timeout: int = 3000,
        compile_memory_limit: int = -1,
        run_memory_limit: int = -1,
    ) -> Dict[str, Any]:
        """
        Execute code using the Piston API.

        Args:
            language: Programming language (e.g., "c", "cpp")
            version: Version of the language (e.g., "10.2.0")
            files: List of files to include in execution (each with "name" and "content")
            stdin: Standard input to provide to the program
            args: Command-line arguments to pass to the program
            compile_timeout: Maximum compilation time in milliseconds
            run_timeout: Maximum execution time in milliseconds
            compile_memory_limit: Maximum memory for compilation in bytes (-1 for unlimited)
            run_memory_limit: Maximum memory for execution in bytes (-1 for unlimited)

        Returns:
            Dictionary containing the execution results
        """
        payload = {
            "language": language,
            "version": version,
            "files": files,
            "stdin": stdin,
            "args": args,
            "compile_timeout": compile_timeout,
            "run_timeout": run_timeout,
            "compile_memory_limit": compile_memory_limit,
            "run_memory_limit": run_memory_limit,
        }

        async with self.session.post(
            f"{self.base_endpoint}/execute",
            json=payload,
            headers={"Content-Type": "application/json"},
        ) as response:
            if response.status != 200:
                error_text = await response.text()
                raise PistonError(
                    f"Error executing code: {response.status} - {error_text}"
                )

            result = await response.json()

            if "message" in result:
                raise PistonError(result["message"])

            return result


def get_piston_client(endpoint: Optional[str] = None) -> PistonClient:
    """
    Get a Piston client instance.

    Args:
        endpoint: Optional custom Piston API endpoint

    Returns:
        PistonClient instance
    """
    piston_endpoint = endpoint or os.environ.get(
        "PISTON_ENDPOINT", "https://emkc.org/api/v2/piston"
    )
    assert isinstance(piston_endpoint, str)
    return PistonClient(base_endpoint=piston_endpoint)


def extract_code_blocks(text: str, language: str = "cpp") -> List[Dict[str, str]]:
    """
    Extract code blocks from text.

    Args:
        text: The text to extract code blocks from
        language: Language to filter by (e.g., "cpp", "c")

    Returns:
        List of dictionaries with "code" and "language" keys
    """
    pattern = r"```(\w*)\n([\s\S]*?)\n```"
    matches = re.findall(pattern, text)

    code_blocks = []
    for lang, code in matches:
        lang = lang.lower()

        if language and lang:
            if language == "cpp" and lang not in ["cpp", "c++"]:
                continue
            elif language == "c" and lang != "c":
                continue
            elif language not in ["c", "cpp"] and language != lang:
                continue

        detected_lang = lang if lang else "unknown"
        code_blocks.append({"language": detected_lang, "code": code.strip()})

    return code_blocks


def add_cpp_includes(code: str) -> str:
    """
    Add common C++ includes if they're missing.

    Args:
        code: C++ code

    Returns:
        Code with added includes if necessary
    """
    if not code:
        return code

    includes = []

    if "#include <iostream>" not in code:
        includes.append("#include <iostream>")
    if "#include <vector>" not in code:
        includes.append("#include <vector>")
    if "#include <string>" not in code:
        includes.append("#include <string>")
    if "#include <bits/stdc++.h>" not in code:
        includes.append("#include <bits/stdc++.h>")
    if "using namespace std;" not in code and "std::" not in code:
        includes.append("using namespace std;")

    if includes:
        return "\n".join(includes) + "\n\n" + code

    return code


def add_c_includes(code: str) -> str:
    """
    Add common C includes if they're missing.

    Args:
        code: C code

    Returns:
        Code with added includes if necessary
    """
    if not code:
        return code

    includes = []

    if "#include <stdio.h>" not in code:
        includes.append("#include <stdio.h>")
    if "#include <stdlib.h>" not in code:
        includes.append("#include <stdlib.h>")
    if "#include <string.h>" not in code:
        includes.append("#include <string.h>")

    if includes:
        return "\n".join(includes) + "\n\n" + code

    return code


async def execute_cpp_code(
    code: str,
    stdin: str = "",
    language: str = "cpp",
    version: str = "11.4.0",
    timeout: int = 5000,
    memory_limit: int = 512000000,
    piston_endpoint: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Execute C/C++ code using the Piston API.

    Args:
        code: C/C++ code to execute
        stdin: Standard input to provide to the program
        language: "c" or "cpp"
        version: Version of the compiler to use
        timeout: Maximum execution time in milliseconds
        memory_limit: Maximum memory in bytes
        piston_endpoint: Optional custom Piston API endpoint

    Returns:
        Dictionary with execution results
    """
    if language == "cpp":
        code = add_cpp_includes(code)
    else:
        code = add_c_includes(code)

    client = get_piston_client(piston_endpoint)

    try:
        main_file = {
            "name": "main.cpp" if language == "cpp" else "main.c",
            "content": code,
        }

        result = await client.execute(
            language=language,
            version=version,
            files=[main_file],
            stdin=stdin,
            compile_timeout=timeout,
            run_timeout=timeout,
            run_memory_limit=memory_limit,
        )

        if "compile" in result and result["compile"]["code"] != 0:
            return {
                "success": False,
                "output": None,
                "error": f"Compilation error: {result['compile']['stderr']}",
            }

        if "run" in result:
            if result["run"]["code"] == 0:
                return {
                    "success": True,
                    "output": result["run"]["stdout"],
                    "error": None,
                }
            else:
                return {
                    "success": False,
                    "output": (
                        result["run"]["stdout"] if result["run"]["stdout"] else None
                    ),
                    "error": f"Runtime error (exit code {result['run']['code']}): {result['run']['stderr']}",
                }

        return {
            "success": False,
            "output": None,
            "error": "Unknown error during execution",
        }

    except PistonError as e:
        return {
            "success": False,
            "output": None,
            "error": f"Piston error: {str(e)}",
        }
    except Exception as e:
        return {"success": False, "output": None, "error": f"Error: {str(e)}"}
    finally:
        loop = asyncio.get_event_loop()
        loop.create_task(client.close())


def compare_outputs(actual: str, expected: str) -> float:
    """
    Compare actual and expected outputs to calculate a similarity score.

    Args:
        actual: Actual output from code execution
        expected: Expected output

    Returns:
        Similarity score between 0.0 and 1.0
    """
    if actual is None:
        actual = ""
    if expected is None:
        expected = ""

    actual_norm = re.sub(r"\s+", " ", actual.strip())
    expected_norm = re.sub(r"\s+", " ", expected.strip())

    if actual_norm == expected_norm:
        return 1.0

    try:
        actual_num = float(actual_norm)
        expected_num = float(expected_norm)

        if expected_num == 0:
            return 1.0 if actual_num == 0 else 0.0

        rel_diff = abs(actual_num - expected_num) / abs(expected_num)

        if rel_diff <= 0.001:
            return 1.0
        elif rel_diff <= 0.01:
            return 0.95
        elif rel_diff <= 0.1:
            return 0.7
        else:
            return max(0.0, 1.0 - min(1.0, rel_diff))
    except (ValueError, TypeError):
        pass

    if "\n" in actual_norm or "\n" in expected_norm:
        actual_lines = actual_norm.split("\n")
        expected_lines = expected_norm.split("\n")

        common_len = min(len(actual_lines), len(expected_lines))
        if common_len == 0:
            return 0.0

        line_similarities = []
        for i in range(common_len):
            if actual_lines[i] == expected_lines[i]:
                line_similarities.append(1.0)
            else:
                line_similarities.append(
                    string_similarity(actual_lines[i], expected_lines[i])
                )

        total_weight = sum(1 / (i + 1) for i in range(common_len))
        weighted_sum = sum(
            (1 / (i + 1)) * sim for i, sim in enumerate(line_similarities)
        )
        similarity = weighted_sum / total_weight if total_weight > 0 else 0.0

        length_penalty = min(len(actual_lines), len(expected_lines)) / max(
            len(actual_lines), len(expected_lines)
        )

        return similarity * length_penalty

    return string_similarity(actual_norm, expected_norm)


def string_similarity(s1: str, s2: str) -> float:
    """
    Calculate string similarity.

    Args:
        s1: First string
        s2: Second string

    Returns:
        Similarity score between 0.0 and 1.0
    """
    if not s1 and not s2:
        return 1.0
    if not s1 or not s2:
        return 0.0

    distance = levenshtein_distance(s1, s2)
    max_len = max(len(s1), len(s2))
    return 1.0 - (distance / max_len if max_len > 0 else 0.0)


def levenshtein_distance(s1: str, s2: str) -> int:
    """
    Calculate the Levenshtein distance between two strings.

    Args:
        s1: First string
        s2: Second string

    Returns:
        Edit distance between strings
    """
    if len(s1) < len(s2):
        return levenshtein_distance(s2, s1)

    if not s2:
        return len(s1)

    previous_row: List[int] = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


async def run_cpp_test_cases(
    code: str,
    test_cases: List[Dict[str, Any]],
    language: str = "cpp",
    version: str = "11.4.0",
    timeout: int = 5000,
    memory_limit: int = 512000000,
    piston_endpoint: Optional[str] = None,
) -> List[TestResult]:
    """
    Run C/C++ code against multiple test cases.

    Args:
        code: C/C++ code to execute
        test_cases: List of test cases with "input" and "expected_output" keys
        language: "c" or "cpp"
        version: Version of the compiler to use
        timeout: Maximum execution time in milliseconds
        memory_limit: Maximum memory in bytes
        piston_endpoint: Optional custom Piston API endpoint

    Returns:
        List of TestResult objects
    """
    results = []

    for i, test_case in enumerate(test_cases):
        test_input = test_case.get("input", "")
        expected_output = test_case.get("expected_output", "")
        test_name = test_case.get("name", f"Test {i+1}")

        execution_result = await execute_cpp_code(
            code=code,
            stdin=test_input,
            language=language,
            version=version,
            timeout=timeout,
            memory_limit=memory_limit,
            piston_endpoint=piston_endpoint,
        )

        test_result = TestResult(test_name=test_name, expected_output=expected_output)

        if execution_result["success"]:
            actual_output = execution_result["output"]
            test_result.actual_output = actual_output
            similarity = compare_outputs(actual_output, expected_output)
            test_result.score = similarity

            if similarity >= 0.99:
                test_result.status = "AC"
            elif similarity > 0:
                test_result.status = "PA"
            else:
                test_result.status = "WA"
            test_result.feedback = f"Similarity: {similarity:.2f}"
        else:
            test_result.status = (
                "CE" if "Compilation error" in execution_result["error"] else "RE"
            )
            test_result.feedback = execution_result["error"]
            test_result.score = 0.0

        results.append(test_result)
        if test_result.score == 0.0:
            break

    return results


@reward_function
def ioi_cpp_code_reward(
    messages: List[Message],
    ground_truth: Union[Optional[str], Optional[List[Dict[str, Any]]]],
    language: str = "cpp",
    version: str = "11.4.0",
    timeout: int = 5000,
    memory_limit: int = 512000000,
    piston_endpoint: Optional[str] = None,
    pass_threshold: float = 0.99,
    **kwargs: Any,
) -> EvaluateResult:
    """
    Wrapper function for the asynchronous implementation to make it compatible with the reward_function decorator.
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        return _ioi_cpp_code_reward_impl(
            messages=messages,
            ground_truth=ground_truth,
            language=language,
            version=version,
            timeout=timeout,
            memory_limit=memory_limit,
            piston_endpoint=piston_endpoint,
            pass_threshold=pass_threshold,
            **kwargs,
        )
    finally:
        loop.close()


def _ioi_cpp_code_reward_impl(
    messages: List[Message],
    ground_truth: Union[Optional[str], Optional[List[Dict[str, Any]]]],
    language: str = "cpp",
    version: str = "11.4.0",
    timeout: int = 5000,
    memory_limit: int = 512000000,
    piston_endpoint: Optional[str] = None,
    pass_threshold: float = 0.99,
    **kwargs: Any,
) -> EvaluateResult:
    """
    Evaluate C/C++ code correctness using the Piston execution engine.

    This function evaluates code for competitive programming problems (like IOI)
    by compiling and executing C/C++ code against test cases.

    Args:
        messages: Generated conversation messages
        ground_truth: Expected output string or list of test case dictionaries.
        language: Programming language ("c" or "cpp")
        version: Version of the compiler to use
        timeout: Maximum execution time in milliseconds
        memory_limit: Maximum memory in bytes
        piston_endpoint: Optional custom Piston API endpoint
        pass_threshold: Similarity threshold for considering a test passed
        **kwargs: Additional keyword arguments

    Returns:
        EvaluateResult with score and metrics
    """
    metrics: Dict[str, MetricResult] = {}

    if (
        not messages
        or not isinstance(messages[-1], Message)
        or messages[-1].role != "assistant"
        or messages[-1].content is None
    ):
        return EvaluateResult(
            score=0.0,
            reason="Invalid or missing assistant response in messages.",
            metrics={
                "error": MetricResult(
                    score=0.0,
                    is_score_valid=False,
                    reason="Last message not a valid assistant response.",
                )
            },
        )

    response_content = messages[-1].content

    expected_output_str_from_gt: Optional[str] = None
    test_cases_from_gt: Optional[List[Dict[str, Any]]] = None

    if isinstance(ground_truth, str):
        expected_output_str_from_gt = ground_truth
    elif isinstance(ground_truth, list):
        if all(isinstance(item, dict) for item in ground_truth):
            test_cases_from_gt = ground_truth
        else:
            return EvaluateResult(
                score=0.0,
                reason="Invalid ground_truth format: if list, must be list of test case dicts.",
                metrics={
                    "error": MetricResult(
                        score=0.0,
                        is_score_valid=False,
                        reason="Invalid ground_truth list format.",
                    )
                },
            )
    elif ground_truth is not None:
        return EvaluateResult(
            score=0.0,
            reason="Invalid ground_truth format: expected string, list of test case dicts, or None.",
            metrics={
                "error": MetricResult(
                    score=0.0,
                    is_score_valid=False,
                    reason="Invalid ground_truth format.",
                )
            },
        )

    code_blocks = extract_code_blocks(response_content, language)

    if not code_blocks:
        return EvaluateResult(
            score=0.0,
            reason=f"No {language} code blocks found in model's response.",
            metrics={
                "error": MetricResult(
                    score=0.0,
                    is_score_valid=False,
                    reason=f"No {language} code blocks found in model's response.",
                )
            },
        )

    code = code_blocks[0]["code"]

    metrics["extracted_code"] = MetricResult(
        score=0.0,
        is_score_valid=True,
        reason=f"Extracted code:\n```{language}\n{code}\n```",
    )

    if expected_output_str_from_gt and not test_cases_from_gt:
        metrics["expected_output"] = MetricResult(
            score=0.0,
            is_score_valid=True,
            reason=f"Expected output:\n{expected_output_str_from_gt}",
        )

    if test_cases_from_gt:
        results = asyncio.get_event_loop().run_until_complete(
            run_cpp_test_cases(
                code=code,
                test_cases=test_cases_from_gt,
                language=language,
                version=version,
                timeout=timeout,
                memory_limit=memory_limit,
                piston_endpoint=piston_endpoint,
            )
        )

        passed = sum(1 for result in results if result.score >= pass_threshold)
        total = len(results)
        overall_score = passed / total if total > 0 else 0.0
        final_reason = f"{passed}/{total} tests passed ({overall_score:.2%})."

        metrics["test_results"] = MetricResult(
            score=overall_score,
            is_score_valid=overall_score >= pass_threshold,
            reason=json.dumps(
                [
                    {
                        "test_name": result.test_name,
                        "status": result.status,
                        "score": result.score,
                        "feedback": result.feedback,
                    }
                    for result in results
                ],
                indent=2,
            ),
        )

        metrics["pass_rate"] = MetricResult(
            score=overall_score,
            is_score_valid=overall_score == 1.0,
            reason=f"{passed}/{total} tests passed ({overall_score:.2%})",
        )

        return EvaluateResult(score=overall_score, reason=final_reason, metrics=metrics)

    elif expected_output_str_from_gt:
        execution_result = asyncio.get_event_loop().run_until_complete(
            execute_cpp_code(
                code=code,
                language=language,
                version=version,
                timeout=timeout,
                memory_limit=memory_limit,
                piston_endpoint=piston_endpoint,
            )
        )

        if execution_result["success"]:
            output = execution_result["output"]
            final_reason = "Code executed successfully."

            metrics["execution_result"] = MetricResult(
                score=1.0,
                is_score_valid=True,
                reason=f"Code executed successfully with output:\n{output}",
            )

            similarity = compare_outputs(output, expected_output_str_from_gt)
            match_reason = f"Output similarity: {similarity:.2f}\n\nExpected:\n{expected_output_str_from_gt}\n\nActual:\n{output}"
            final_reason += f" Output similarity: {similarity:.2f}."

            metrics["output_match"] = MetricResult(
                score=similarity,
                is_score_valid=similarity >= pass_threshold,
                reason=match_reason,
            )

            return EvaluateResult(
                score=similarity, reason=final_reason, metrics=metrics
            )
        else:
            error = execution_result["error"]
            final_reason = f"Code execution failed: {error}"

            metrics["execution_result"] = MetricResult(
                score=0.0,
                is_score_valid=False,
                reason=f"Code execution failed with error:\n{error}",
            )

            return EvaluateResult(score=0.0, reason=final_reason, metrics=metrics)

    else:
        execution_result = asyncio.get_event_loop().run_until_complete(
            execute_cpp_code(
                code=code,
                language=language,
                version=version,
                timeout=timeout,
                memory_limit=memory_limit,
                piston_endpoint=piston_endpoint,
            )
        )

        if execution_result["success"]:
            output = execution_result["output"]
            final_reason = (
                "Code executed successfully (no expected output for comparison)."
            )

            metrics["execution_result"] = MetricResult(
                score=1.0,
                is_score_valid=True,
                reason=f"Code executed successfully with output:\n{output}",
            )

            return EvaluateResult(score=1.0, reason=final_reason, metrics=metrics)
        else:
            error = execution_result["error"]
            final_reason = f"Code execution failed: {error}"
            metrics["execution_result"] = MetricResult(
                score=0.0,
                is_score_valid=False,
                reason=f"Code execution failed with error:\n{error}",
            )

            return EvaluateResult(score=0.0, reason=final_reason, metrics=metrics)


@reward_function
def binary_cpp_code_reward(
    messages: List[Message],
    ground_truth: Union[Optional[str], Optional[List[Dict[str, Any]]]],
    language: str = "cpp",
    version: str = "11.4.0",
    timeout: int = 5000,
    memory_limit: int = 512000000,
    piston_endpoint: Optional[str] = None,
    pass_threshold: float = 0.99,
    **kwargs: Any,
) -> EvaluateResult:
    """
    Evaluate C/C++ code correctness and return a binary result (passed/failed).

    This function is a wrapper around ioi_cpp_code_reward that returns 1.0 if the
    score is at or above the pass_threshold, and 0.0 otherwise.

    Args:
        messages: Generated conversation messages
        ground_truth: Expected output string or list of test case dictionaries.
        language: Programming language ("c" or "cpp")
        version: Version of the compiler to use
        timeout: Maximum execution time in milliseconds
        memory_limit: Maximum memory in bytes
        piston_endpoint: Optional custom Piston API endpoint
        pass_threshold: Similarity threshold for considering a test passed
        **kwargs: Additional keyword arguments

    Returns:
        EvaluateResult with binary score (0.0 or 1.0) and metrics
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        reward_output = _ioi_cpp_code_reward_impl(
            messages=messages,
            ground_truth=ground_truth,
            language=language,
            version=version,
            timeout=timeout,
            memory_limit=memory_limit,
            piston_endpoint=piston_endpoint,
            pass_threshold=pass_threshold,
            **kwargs,
        )

        score = reward_output.score
        binary_score = 1.0 if score >= pass_threshold else 0.0
        metrics = dict(reward_output.metrics)
        final_reason = f"Binary score based on threshold {pass_threshold:.2f}. Original score: {score:.2f}."
        metrics["binary_result"] = MetricResult(
            score=binary_score,
            is_score_valid=binary_score == 1.0,
            reason=f"{'Passed' if binary_score > 0 else 'Failed'} (threshold: {pass_threshold:.2f}, actual: {score:.2f})",
        )

        return EvaluateResult(score=binary_score, reason=final_reason, metrics=metrics)
    finally:
        loop.close()



import json
import os
import sys
from unittest.mock import MagicMock, patch

import aiohttp
import pytest

try:
    import torch  # type: ignore
except Exception:  # pragma: no cover - optional dependency
    torch = None

# Ensure reward-kit is in the path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from reward_kit.models import EvaluateResult, Message  # Removed PreviewBulk* models
from reward_kit.rewards.math import math_reward

# Import functions from the example scripts if they are structured for import
# For simplicity here, we might re-implement small parts or directly call reward functions


# --- Fixtures ---


@pytest.fixture
def mock_fireworks_api_key(monkeypatch):
    monkeypatch.setenv("FIREWORKS_API_KEY", "test_api_key_for_readiness_tests")


@pytest.fixture
def mock_requests_post():
    with patch("requests.post") as mock_post:
        yield mock_post


# To run these tests: pytest tests/test_readiness.py -s (to see print statements)
# The -s flag is helpful for seeing the script outputs during test runs.

import subprocess


# --- End-to-End Script Tests for Math Example ---
class TestMathExampleEndToEndScripts:

    BASE_MATH_EXAMPLE_PATH = os.path.join(
        os.path.dirname(__file__), "../examples/math_example"
    )

    def run_script(
        self, script_name: str, env_vars: dict = None, timeout_seconds: int = 180
    ) -> subprocess.CompletedProcess:
        """Helper to run an example script."""
        script_path = os.path.join(self.BASE_MATH_EXAMPLE_PATH, script_name)
        command = [
            sys.executable,
            script_path,
        ]  # Use sys.executable to ensure correct python version

        current_env = os.environ.copy()
        project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

        existing_pythonpath = current_env.get("PYTHONPATH")
        if existing_pythonpath:
            current_env["PYTHONPATH"] = (
                f"{project_root}{os.pathsep}{existing_pythonpath}"
            )
        else:
            current_env["PYTHONPATH"] = project_root

        if env_vars:
            current_env.update(env_vars)

        process = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=self.BASE_MATH_EXAMPLE_PATH,  # Run script from its directory
            env=current_env,
            timeout=timeout_seconds,
        )
        print(f"\n--- Output for {script_name} (timeout: {timeout_seconds}s) ---")
        print(f"STDOUT:\n{process.stdout}")
        if process.stderr:
            print(f"STDERR:\n{process.stderr}")
        print("--- End Output ---")
        return process

    # @pytest.mark.skipif(
    #     os.environ.get("CI") == "true",
    #     reason="Skipping resource-intensive TRL integration test in CI",
    # )
    # @pytest.mark.timeout(
    #     630
    # )  # Timeout for test function (slightly > subprocess timeout)
    # @patch("trl.GRPOTrainer")
    # @patch("peft.get_peft_model")
    # @patch("transformers.AutoModelForCausalLM.from_pretrained")
    # @patch("transformers.AutoTokenizer.from_pretrained")
    # @patch("datasets.Dataset.from_list")  # Mock dataset loading
    # @patch("datasets.Dataset.map")  # Mock dataset map where it's used
    # def test_e2e_trl_grpo_integration_script(
    #     self,
    #     mock_dataset_map,  # New mock
    #     mock_dataset_from_list,  # New mock
    #     mock_tokenizer_load,
    #     mock_base_model_load,
    #     mock_get_peft_model,
    #     mock_grpo_trainer_class,
    # ):
    #     """End-to-end test for examples/math_example/trl_grpo_integration.py with mocked TRL steps."""
    #     print("\nRunning E2E Test: Math Example - trl_grpo_integration.py (Mocked TRL)")

    #     # Configure mocks
    #     mock_tokenizer = MagicMock()
    #     mock_tokenizer.pad_token = "<|endoftext|>"
    #     mock_tokenizer.eos_token_id = 50256
    #     mock_tokenizer_load.return_value = mock_tokenizer

    #     mock_base_model = MagicMock()  # Mock for the base model
    #     mock_base_model_load.return_value = mock_base_model

    #     mock_peft_model = MagicMock()
    #     mock_peft_model.print_trainable_parameters = MagicMock()
    #     mock_peft_model.device = torch.device("cpu")  # Add device attribute
    #     mock_get_peft_model.return_value = mock_peft_model

    #     # Configure dataset mocks
    #     mock_mapped_dataset = MagicMock()
    #     mock_mapped_dataset.set_format = MagicMock()
    #     mock_dataset_map.return_value = (
    #         mock_mapped_dataset  # mock_dataset_map is the mock for dataset_instance.map
    #     )

    #     mock_dataset_instance = MagicMock()
    #     mock_dataset_instance.map = (
    #         mock_dataset_map  # Assign the .map mock to the instance
    #     )
    #     mock_dataset_from_list.return_value = (
    #         mock_dataset_instance  # Dataset.from_list returns this instance
    #     )

    #     # Configure the instance returned by the mocked GRPOTrainer class
    #     mock_grpo_trainer_instance = MagicMock()
    #     mock_grpo_trainer_instance.step.return_value = {"loss": 0.1, "reward": 0.9}

    #     # Mock the dataloader and accelerator more completely
    #     mock_dataloader = MagicMock()
    #     # Ensure batch tensors are on the same device the trainer expects
    #     mock_batch_input_ids = torch.randint(0, 100, (1, 10), device="cpu")
    #     mock_batch = {
    #         "input_ids": mock_batch_input_ids,
    #         "query": ["mock query"],
    #         "response": ["mock response"],
    #     }
    #     mock_dataloader.__iter__.return_value = iter([mock_batch])
    #     # mock_grpo_trainer_instance.dataloader = mock_dataloader # No longer needed directly
    #     mock_grpo_trainer_instance.get_train_dataloader = MagicMock(
    #         return_value=mock_dataloader
    #     )  # Mock get_train_dataloader

    #     mock_accelerator = MagicMock()
    #     mock_accelerator.device = torch.device("cpu")
    #     mock_grpo_trainer_instance.accelerator = mock_accelerator

    #     # Mock generate method if called by step or before
    #     mock_grpo_trainer_instance.generate = MagicMock(
    #         return_value=torch.randint(0, 100, (1, 5), device="cpu")
    #     )

    #     mock_grpo_trainer_class.return_value = mock_grpo_trainer_instance

    #     env_vars = {"TEST_MODE_TRL": "true"}
    #     # Run the script with a 10-minute (600 seconds) timeout
    #     result = self.run_script(
    #         "trl_grpo_integration.py", env_vars=env_vars, timeout_seconds=600
    #     )

    #     assert (
    #         result.returncode == 0
    #     ), f"trl_grpo_integration.py script failed with exit code {result.returncode}. Stderr: {result.stderr}"
    #     assert (
    #         "GRPO training loop completed for Math Example." in result.stdout
    #     ), "Expected completion message not found in trl_grpo_integration.py output."

    #     from examples.math_example.trl_grpo_integration import (
    #         grpo_config as math_grpo_config,
    #     )

    #     # The script now calls grpo_trainer.train(), not grpo_trainer.step() directly.
    #     # The mock_grpo_trainer_instance.train method is not called because the script runs in a subprocess
    #     # where the @patch decorator does not apply.
    #     # The assertions on result.returncode and stdout content are the primary checks for this E2E script test.
    #     # mock_grpo_trainer_instance.train.assert_called_once() # This line is removed.
    #     # The number of steps taken internally by train() will be 1 due to TEST_MODE_TRL=true in the script's env.
    #     # We can't easily check internal step calls on the mock of GRPOTrainer itself when train() is called.
    #     # The script's output "GRPO training loop completed for Math Example." and return code 0 are primary indicators.
    #     # The assertion on result.returncode == 0 and the completion message in stdout already cover this.
    #     # If we wanted to check logs for number of steps, that would be parsing stdout.
    #     # For now, asserting train() was called is the most direct check on the mock.

    #     print("E2E Test: Math Example - trl_grpo_integration.py (Mocked TRL): PASSED")


# --- End-to-End Script Tests for Math Example (OpenR1) ---
class TestMathExampleOpenR1EndToEndScripts:

    BASE_MATH_EXAMPLE_OPENR1_PATH = os.path.join(
        os.path.dirname(__file__), "../examples/math_example_openr1"
    )

    def run_script(
        self, script_name: str, env_vars: dict = None, timeout_seconds: int = 180
    ) -> subprocess.CompletedProcess:
        """Helper to run an example script for OpenR1."""
        script_path = os.path.join(self.BASE_MATH_EXAMPLE_OPENR1_PATH, script_name)
        command = [
            sys.executable,
            script_path,
        ]

        current_env = os.environ.copy()
        project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

        existing_pythonpath = current_env.get("PYTHONPATH")
        if existing_pythonpath:
            current_env["PYTHONPATH"] = (
                f"{project_root}{os.pathsep}{existing_pythonpath}"
            )
        else:
            current_env["PYTHONPATH"] = project_root

        if env_vars:
            current_env.update(env_vars)

        process = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=self.BASE_MATH_EXAMPLE_OPENR1_PATH,  # Run script from its directory
            env=current_env,
            timeout=timeout_seconds,
        )
        print(
            f"\n--- Output for {script_name} (OpenR1, timeout: {timeout_seconds}s) ---"
        )
        print(f"STDOUT:\n{process.stdout}")
        if process.stderr:
            print(f"STDERR:\n{process.stderr}")
        print("--- End Output ---")
        return process

    # @pytest.mark.skipif(
    #     os.environ.get("CI") == "true",
    #     reason="Skipping resource-intensive TRL integration test in CI",
    # )
    # @patch("trl.GRPOTrainer")
    # @patch("peft.get_peft_model")
    # @patch("transformers.AutoModelForCausalLM.from_pretrained")
    # @patch("transformers.AutoTokenizer.from_pretrained")
    # @patch("datasets.Dataset.from_list")
    # @patch("datasets.Dataset.map")
    # def test_e2e_trl_grpo_integration_script_openr1(
    #     self,
    #     mock_dataset_map_openr1,
    #     mock_dataset_from_list_openr1,
    #     mock_tokenizer_load_openr1,
    #     mock_base_model_load_openr1,
    #     mock_get_peft_model_openr1,
    #     mock_grpo_trainer_class_openr1,
    # ):
    #     """End-to-end test for examples/math_example_openr1/trl_grpo_integration.py with mocked TRL steps."""
    #     print(
    #         "\nRunning E2E Test: Math Example OpenR1 - trl_grpo_integration.py (Mocked TRL)"
    #     )

    #     # Configure mocks
    #     mock_tokenizer = MagicMock()
    #     mock_tokenizer.pad_token = "<|endoftext|>"
    #     mock_tokenizer.eos_token_id = (
    #         50256  # Example, ensure it matches model if relevant
    #     )
    #     mock_tokenizer_load_openr1.return_value = mock_tokenizer

    #     mock_base_model = MagicMock()
    #     mock_base_model_load_openr1.return_value = mock_base_model

    #     mock_peft_model = MagicMock()
    #     mock_peft_model.print_trainable_parameters = MagicMock()
    #     mock_peft_model.device = torch.device("cpu")
    #     mock_get_peft_model_openr1.return_value = mock_peft_model

    #     mock_mapped_dataset = MagicMock()
    #     mock_mapped_dataset.set_format = MagicMock()
    #     mock_dataset_map_openr1.return_value = mock_mapped_dataset

    #     mock_dataset_instance = MagicMock()
    #     mock_dataset_instance.map = mock_dataset_map_openr1
    #     mock_dataset_from_list_openr1.return_value = mock_dataset_instance

    #     mock_grpo_trainer_instance = MagicMock()
    #     mock_grpo_trainer_instance.train = MagicMock()  # Mock the train method directly

    #     # Mock dataloader and accelerator parts if GRPOTrainer's train() needs them internally from the instance
    #     mock_dataloader = MagicMock()
    #     mock_batch_input_ids = torch.randint(0, 100, (1, 10), device="cpu")
    #     mock_batch = {
    #         "input_ids": mock_batch_input_ids,
    #         "query": ["mock query openr1"],
    #         "response": ["mock response openr1"],
    #     }
    #     mock_dataloader.__iter__.return_value = iter([mock_batch])
    #     mock_grpo_trainer_instance.get_train_dataloader = MagicMock(
    #         return_value=mock_dataloader
    #     )

    #     mock_accelerator = MagicMock()
    #     mock_accelerator.device = torch.device("cpu")
    #     mock_grpo_trainer_instance.accelerator = mock_accelerator

    #     mock_grpo_trainer_class_openr1.return_value = mock_grpo_trainer_instance

    #     env_vars = {"TEST_MODE_TRL": "true"}
    #     result = self.run_script(
    #         "trl_grpo_integration.py", env_vars=env_vars, timeout_seconds=600
    #     )  # Increased timeout

    #     assert (
    #         result.returncode == 0
    #     ), f"OpenR1 trl_grpo_integration.py script failed with exit code {result.returncode}. Stderr: {result.stderr}"
    #     assert (
    #         "GRPO training loop completed for OpenR1 Math Example." in result.stdout
    #     ), "Expected completion message not found in OpenR1 trl_grpo_integration.py output."

    #     # Since the script runs in a subprocess, the mocks apply to the script's execution context if it imports them.
    #     # The primary check is the script's output and return code.
    #     # We can't directly assert mock_grpo_trainer_instance.train.assert_called_once() here
    #     # because the mock object `mock_grpo_trainer_instance` is in the test process,
    #     # not the subprocess where the script ran.

    #     print(
    #         "E2E Test: Math Example OpenR1 - trl_grpo_integration.py (Mocked TRL): PASSED"
    #     )



"""
Tests for math reward functions.
"""

import re
from typing import List, Optional, Tuple, Union  # Added Tuple

import pytest

# Removed: from reward_kit.rewards.advanced_math import advanced_math_reward
from reward_kit.models import Message  # Added Message import
from reward_kit.models import EvaluateResult, MetricResult
from reward_kit.rewards.math import compare_numbers, extract_numbers, math_reward


class TestExtractNumbers:
    # --- PRIORITY 1: Boxed LaTeX ---
    def test_P1_boxed_simple_number(self):
        text = "The answer is \\boxed{42}."
        expected = [("\\boxed{42}", 42.0)]
        assert extract_numbers(text) == expected

    def test_P1_boxed_simple_fraction_string(self):
        text = "The answer is \\boxed{1/2}."
        expected = [("\\boxed{1/2}", 0.5)]
        assert extract_numbers(text) == expected

    def test_P1_boxed_latex_fraction(self):
        text = "The answer is \\boxed{\\frac{3}{4}}."
        expected = [("\\boxed{\\frac{3}{4}}", 0.75)]
        assert extract_numbers(text) == expected

    def test_P1_boxed_mcq_letter(self):
        text = "The answer is \\boxed{A}."
        expected = [("\\boxed{A}", "A")]
        assert extract_numbers(text) == expected

    def test_P1_boxed_or_expression(self):
        text = "The roots are \\boxed{x=1 \\text{ or } x=2}."
        expected = [("\\boxed{x=1 \\text{ or } x=2}", "x=1 \\text{ or } x=2")]
        assert extract_numbers(text) == expected

    def test_P1_boxed_complex_content_as_string_or_ignored(self):
        text_no_or = "The answer is \\boxed{x^2+1}."
        # P1: found_any_boxed_expr = True. boxed_answers is empty. Returns []. Correct.
        assert extract_numbers(text_no_or) == []

        text_with_or = "The answer is \\boxed{x^2+1 \\text{ or } y=0}."
        expected_with_or = [
            ("\\boxed{x^2+1 \\text{ or } y=0}", "x^2+1 \\text{ or } y=0")
        ]
        assert extract_numbers(text_with_or) == expected_with_or

    def test_P1_multiple_boxed_expressions(self):
        text = "Answers: \\boxed{1}, then \\boxed{B}, and finally \\boxed{1/2 or 3/2}"
        # Order of extraction depends on finditer, so check presence and content
        results = extract_numbers(text)
        assert len(results) == 3
        assert ("\\boxed{1}", 1.0) in results
        assert ("\\boxed{B}", "B") in results
        assert ("\\boxed{1/2 or 3/2}", "1/2 or 3/2") in results

    # --- PRIORITY 2: GSM8K-style (#### ...) ---
    def test_P2_gsm8k_simple_number(self):
        text = "The final answer is #### 123."  # Input with period
        expected = [
            ("#### 123", 123.0)
        ]  # Regex for num content won't include trailing period in m.group(0)
        assert extract_numbers(text) == expected

    def test_P2_gsm8k_with_commas_and_decimal(self):
        text = "The final answer is #### 1,234.56."  # Input with period
        expected = [
            ("#### 1,234.56", 1234.56)
        ]  # Expect m.group(0) to be "#### 1,234.56"
        assert extract_numbers(text) == expected

    def test_P2_gsm8k_negative_number(self):
        text = "The final answer is #### -7."  # Input with period
        expected = [("#### -7", -7.0)]
        assert extract_numbers(text) == expected

    def test_P2_gsm8k_takes_priority_over_mcq_and_general(self):
        text = "Answer is (A). The number is 10. Final: #### 100."  # Input with period
        expected = [("#### 100", 100.0)]
        assert extract_numbers(text) == expected

    def test_P2_multiple_gsm8k_markers(self):
        text = "First #### 1. Second #### 2."  # Input with periods
        results = extract_numbers(text)
        assert len(results) == 2
        assert ("#### 1", 1.0) in results
        assert ("#### 2", 2.0) in results

    # --- PRIORITY 3: Multiple Choice Question (MCQ) ---
    # MCQ extraction is now handled by multiple_choice_math_reward.py
    # These tests verify that extract_numbers from math.py NO LONGER extracts MCQs.
    def test_P3_mcq_various_formats(self):
        assert extract_numbers("(A)") == []
        assert extract_numbers("B.") == []
        assert extract_numbers("[C]") == []
        assert extract_numbers("{D}") == []
        assert extract_numbers(" E ") == []

    def test_P3_mcq_multiple_different_mcqs(self):
        text = "Options: (A), B., [C]"
        results = extract_numbers(text)
        assert results == []

    def test_P3_mcq_takes_priority_over_general_numbers(self):
        # Since MCQ is no longer extracted by math.extract_numbers, general numbers should be found.
        text = "The number is 10, but the choice is (A)."
        expected = [("10", 10.0)]
        assert extract_numbers(text) == expected

    # --- PRIORITY 4: General Fallback Extraction ---
    def test_P4_general_integers(self):
        text = "The answer is 42. Another value is -17."
        results = extract_numbers(text)
        assert ("42", 42.0) in results
        assert ("-17", -17.0) in results
        assert len(results) == 2

    def test_P4_general_decimals(self):
        text = "The value of pi is 3.14159."
        expected = [("3.14159", 3.14159)]
        assert extract_numbers(text) == expected

    def test_P4_general_scientific_notation(self):
        text = "Avogadro's number is approximately 6.022e23."
        expected = [("6.022e23", 6.022e23)]
        assert extract_numbers(text) == expected

    def test_P4_general_fractions_plain(self):
        text = "One half is 1/2 and three quarters is 3/4."
        results = extract_numbers(text)
        # m.group(0) for "1/2 and" with the current frac_pattern should be "1/2"
        assert ("1/2", 0.5) in results
        assert ("3/4", 0.75) in results
        assert len(results) == 2

    def test_P4_general_latex_fraction_outside_box(self):
        text = "Answer: $\\frac{1}{2}$"
        expected = [("\\frac{1}{2}", 0.5)]
        assert extract_numbers(text) == expected

    def test_P4_general_latex_sci_notation_outside_box(self):
        text = "Value: $3 \\times 10^{8}$"
        expected = [("3 \\times 10^{8}", 3e8)]
        assert extract_numbers(text) == expected

    def test_P4_general_numbers_with_units_no_algebraic_vars(self):
        text = "The distance is 42 km and the weight is 3.5 kg."
        results = extract_numbers(text)
        # Order might vary, check presence
        assert ("42 km", 42.0) in results
        assert ("3.5 kg", 3.5) in results
        assert len(results) == 2

    def test_P4_general_avoids_coefficients_plain_text(self):
        assert extract_numbers("4y") == []
        assert extract_numbers("4 y") == []
        results = extract_numbers("answer = 4x + 2")
        assert len(results) == 1
        assert results[0] == ("2", 2.0)  # Only "2" should be extracted

    def test_P4_general_avoids_coefficients_latex(self):
        assert extract_numbers("$4y$") == []
        assert extract_numbers("$4 y$") == []
        assert extract_numbers("$x=4a+2b$") == []
        results_eq = extract_numbers("$x=4$")  # "4" is extracted from "$x=4$"
        assert len(results_eq) == 1
        assert results_eq[0] == ("4", 4.0)

    def test_P4_general_multiple_formats_fallback(self):
        text = "Values: 42, then 3.14, also $1/4$, and $10 \\text{ m}$, finally $5.5 \\times 10^{6} \\text{ Hz}$"
        results = extract_numbers(text)
        extracted_values = {r[1] for r in results}
        expected_values = {42.0, 3.14, 0.25, 10.0, 5.5e6}
        assert extracted_values == expected_values
        # Check original texts for some to ensure they are reasonable
        # This can be tricky due to overlapping regexes and filtering in fallback.
        # For example, "$10 \\text{ m}$" might extract "10" with original text "10".
        # For "$1/4$", original text should be "1/4".
        # For "$5.5 \\times 10^{6} \\text{ Hz}$", original text "5.5 \\times 10^{6}".
        assert any(r[0] == "1/4" and r[1] == 0.25 for r in results)
        assert any(r[0] == "5.5 \\times 10^{6}" and r[1] == 5.5e6 for r in results)

    # --- Overall Priority Tests ---
    def test_priority_boxed_over_all_else(self):
        text = "The answer is (A), also 10, and #### 20, but finally \\boxed{30}."
        expected = [("\\boxed{30}", 30.0)]
        assert extract_numbers(text) == expected

    def test_priority_gsm8k_over_mcq_and_general(self):
        text = "The answer is (A), also 10, and finally #### 20."  # Period after 20
        expected = [("#### 20", 20.0)]
        assert extract_numbers(text) == expected

    def test_priority_mcq_over_general(self):
        # MCQ logic removed, so general number '10' should be extracted.
        text = "The answer is 10, but choose (A)."
        expected = [("10", 10.0)]
        assert extract_numbers(text) == expected

    def test_no_answer_found(self):
        text = "This is just some text without any clear answer."
        assert extract_numbers(text) == []

    def test_empty_string(self):
        assert extract_numbers("") == []

    def test_extract_numbers_issue_3_scenario_recheck(self):
        text_coeff = "This is $4 y$."
        assert extract_numbers(text_coeff) == []

        text_valid_num_in_latex = "This is $v_R = 4 \\mathrm{km/h}$."
        results_valid = extract_numbers(text_valid_num_in_latex)
        assert len(results_valid) == 1
        assert results_valid[0] == ("4", 4.0)

        text_combined = "Solution with $4 y$ and also $v_R = 4 \\mathrm{km/h}$."
        results_combined = extract_numbers(text_combined)
        assert len(results_combined) == 1
        assert results_combined[0] == ("4", 4.0)

    # --- PRIORITY 0: HTML/XML tag–wrapped numeric answers ---
    def test_P0_html_tag_simple_number(self):
        text = "Some text <answer>77</answer> more text."
        expected = [("<answer>77</answer>", 77.0)]
        assert extract_numbers(text) == expected

    def test_P0_html_tag_with_whitespace(self):
        text = "<answer>   3.14   </answer>"
        expected = [("<answer>   3.14   </answer>", 3.14)]
        assert extract_numbers(text) == expected

    def test_P0_html_tag_string_content(self):
        text = "<answer>Not a number</answer>"
        expected: List[Tuple[str, Union[str, float]]] = []
        assert extract_numbers(text) == expected

    def test_P0_html_tag_multiple_answer_tags(self):
        text = "<answer>1</answer> and <answer>2.5</answer>"
        results = extract_numbers(text)
        assert ("<answer>1</answer>", 1.0) in results
        assert ("<answer>2.5</answer>", 2.5) in results
        assert len(results) == 2

    def test_P0_html_tag_priority_over_general(self):
        text = "The answer is <answer>42</answer> but also 99."
        expected = [("<answer>42</answer>", 42.0)]
        assert extract_numbers(text) == expected


class TestCompareNumbers:
    def test_exact_match(self):
        is_match, similarity = compare_numbers(42.0, 42.0)
        assert is_match is True
        assert similarity == 1.0

    def test_close_match(self):
        is_match, similarity = compare_numbers(3.14159, 3.14, relative_tolerance=0.01)
        assert is_match is True
        assert similarity == 1.0

    def test_not_close_match(self):
        is_match, similarity = compare_numbers(10.0, 11.0, relative_tolerance=0.01)
        assert is_match is False
        assert similarity < 1.0

    def test_very_different(self):
        is_match, similarity = compare_numbers(100.0, 200.0, relative_tolerance=0.01)
        assert is_match is False
        assert similarity == 0.0

    def test_zero_expected(self):
        is_match, similarity = compare_numbers(0.0, 0.00001, absolute_tolerance=0.0001)
        assert is_match is True
        assert similarity == 1.0

        is_match, similarity = compare_numbers(0.0, 0.0001, absolute_tolerance=0.0001)
        assert is_match is True
        assert similarity == 1.0

        is_match, similarity = compare_numbers(0.0, 0.001, absolute_tolerance=0.0001)
        assert is_match is False
        assert similarity < 1.0


class TestMathReward:
    # Helper to create a common prompt
    def _get_prompt(
        self,
    ) -> Optional[List[Message]]:  # Ensure return type matches Optional if used as such
        return [Message(role="user", content="Solve the math problem.")]

    def test_basic_match_boxed(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="It is \\boxed{4}.")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Answer is \\boxed{4}."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 1.0
        # assert result.score == 1.0 # Corrected access

    def test_basic_match_gsm8k(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="The result is #### 4")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Final answer: #### 4"
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 1.0
        # assert result.score == 1.0 # Corrected access

    def test_basic_match_mcq(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="My answer is (A).")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Choice (A)."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # assert result.score == 0.0 # Duplicate line removed
        assert (
            result.reason is not None
            and "Could not extract answers from original message (ground truth)"
            in result.reason
        )

    def test_type_mismatch_mcq_vs_number(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="It is 4.")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "The number is 4."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 1.0
        # assert result.score == 1.0 # Corrected access

    def test_close_match_tolerance(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="Pi is \\boxed{3.14}.")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Pi is \\boxed{3.14159}."
        result = math_reward(
            messages=messages_arg, ground_truth=ground_truth_arg, tolerance=0.01
        )
        assert isinstance(result, EvaluateResult)
        assert result.score == 1.0
        # assert result.score == 1.0 # Corrected access

    def test_wrong_answer_numeric(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="Answer: \\boxed{5}.")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Answer: \\boxed{4}."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score < 0.1
        # assert result.score < 0.1 # Corrected access

    def test_wrong_answer_mcq(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="Choice (B).")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Choice (A)."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # The reason should be about the original message (ground_truth) if it's unparseable
        assert (
            result.reason is not None
            and "Could not extract answers from original message (ground truth)"
            in result.reason
        )

    def test_no_answer_in_generated(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(role="assistant", content="Answer is (A).")
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Answer is \\boxed{1}."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # assert result.score == 0.0 # Duplicate line removed
        assert (
            result.reason is not None
            and "Could not extract answers from generated message" in result.reason
        )

    def test_no_answer_in_original_when_gen_is_unparseable(
        self,
    ):  # Renamed to avoid conflict and be more descriptive
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant", content="I don't know."
        )  # Gen unparseable
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Answer is \\boxed{1}."  # GT parseable
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        assert (
            result.reason is not None
            and "Could not extract answers from generated message" in result.reason
        )

    def test_unparseable_ground_truth(
        self,
    ):  # Renamed from the second test_no_answer_in_original
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant", content="Answer is \\boxed{1}."
        )  # Gen parseable
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "What is it?"  # GT unparseable
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # assert result.score == 0.0 # Duplicate line removed
        assert (
            result.reason is not None
            and "Could not extract answers from original message (ground truth)"
            in result.reason
        )

    # --- Strictness Penalty Tests ---
    def test_penalty_unboxed_or_issue1(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant", content="The answer is $1/2 \\text{ or } 1$."
        )
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "The answer is $\\boxed{1/2}$."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # assert result.score == 0.0 # Duplicate line removed
        assert result.reason is not None and result.reason.startswith(
            "Strictness fail (Issue #1"
        )
        assert (
            result.metrics["strictness_penalty_unboxed_or"].reason
            == "Generated answer offers multiple numeric alternatives with an unboxed 'or' in the raw response."
        )

    def test_no_penalty_if_gen_is_single_boxed_or_expr(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant", content="The answer is $\\boxed{1/2 \\text{ or } 1}$."
        )
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "The answer is $\\boxed{1/2 \\text{ or } 1}$."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 1.0
        # assert result.score == 1.0 # Duplicate line removed
        assert result.reason is None or ("Strictness fail" not in result.reason)

    def test_penalty_ambiguity_issue2(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant",
            content="The equation is $x=\\frac{1}{4}$. Some other numbers are 1, 2, 3.",
        )
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "The answer is $\\boxed{1/4}$."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        # assert result.score == 0.0 # Duplicate line removed
        assert result.reason is not None and result.reason.startswith(
            "Strictness fail (Issue #2"
        )
        assert (
            result.metrics["strictness_penalty_ambiguity"].reason
            == "Ground truth is specific (one answer), but generated answer is ambiguous (multiple answers extracted, even after potential leniency)."
        )

    def test_issue_false_mcq_match_on_v_B(self):
        user_content_str = "## Task B-1.3.\n\nA ship traveling along a river has covered $24 \\mathrm{~km}$ upstream and $28 \\mathrm{~km}$ downstream. For this journey, it took half an hour less than for traveling $30 \\mathrm{~km}$ upstream and $21 \\mathrm{~km}$ downstream, or half an hour more than for traveling $15 \\mathrm{~km}$ upstream and $42 \\mathrm{~km}$ downstream, assuming that both the ship and the river move uniformly.\n\nDetermine the speed of the ship in still water and the speed of the river."
        assistant_content_generated_str = "## Solution.\n\nLet $t$ be the time required for the boat to travel $24 \\mathrm{~km}$ upstream and $28 \\mathrm{~km}$ downstream, $v_{R}$ the speed of the river, and $v_{B}$ the speed of the boat. When the boat is traveling upstream, its speed is $v_{B}-v_{R}$, and when it is traveling downstream, its speed is $v_{B}+v_{R}$.\n\nSince $t=\\frac{s}{v}$, from the given data, we obtain the following system of equations:\n\n$\\left\\{\\begin{array}{l}t=\\frac{24}{v_{B}-v_{R}}+\\frac{28}{v_{B}+v_{R}} \\\\ t+0.5=\\frac{30}{v_{B}-v_{R}}+\\frac{21}{v_{B}+v_{R}} \\\\ t-0.5=\\frac{15}{v_{B}-v_{R}}+\\frac{42}{v_{B}+v_{R}}\\end{array}\\right.$\n\nBy introducing new variables $x=\\frac{3}{v_{B}-v_{R}}, y=\\frac{7}{v_{B}+v_{R}}$, the system transforms into:\n\n$\\left\\{\\begin{array}{l}t=8 x+4 y \\\\ t+0.5=10 x+3 y \\\\ t-0.5=5 x+6 y\\end{array}\\right.$\n\nSubstituting $t$ from the first equation into the remaining two, we get:\n\n$\\left\\{\\begin{array}{l}8 x+4 y+0.5=10 x+3 y \\\\ 8 x+4 y-0.5=5 x+6 y\\end{array}\\right.$\n\n$\\left\\{\\begin{array}{l}2 x-y=0.5 \\\\ 3 x-2 y=0.5\\end{array}\\right.$\n\nThe solution to the last system is (0.5, 0.5). Then we have:\n\n$\\frac{3}{v_{B}-v_{R}}=0.5$, hence, $v_{B}-v_{R}=6 \\mathrm{~and}$\n\n$\\frac{7}{v_{B}+v_{R}}=0.5$, hence, $v_{B}+v_{R}=14$.\n\nThe speed of the river is $v_{R}=4 \\mathrm{~km} / \\mathrm{h}$, and the speed of the boat is $v_{B}=10 \\mathrm{~km} / \\mathrm{h}$.\n\n## Note:\n\nBy substituting $x=\\frac{1}{v_{B}-v_{R}}, y=\\frac{1}{v_{B}+v_{R}} \\mathrm{~and}$ following the same procedure, the initial system transforms into the system $\\left\\{\\begin{array}{l}6 x-7 y=0.5 \\\\ 9 x-14 y=0.5\\end{array}\\right.$\n\nThe solution to this system is $\\left(\\frac{1}{6}, \\frac{1}{14}\\right)$."
        ground_truth_arg = "The speed of the river is $v_R=4 \\mathrm{km/h}$, and the speed of the boat is $v_B=10 \\mathrm{km/h}$."

        user_message = Message(role="user", content=user_content_str)
        assistant_message = Message(
            role="assistant", content=assistant_content_generated_str
        )
        messages_arg = [user_message, assistant_message]

        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        extracted_gen_reason = result.metrics["extracted_generated_answers"].reason
        extracted_orig_reason = result.metrics["extracted_original_answers"].reason
        assert "'B'" not in extracted_gen_reason
        assert "'B'" not in extracted_orig_reason
        assert result.score == 0.0
        assert (
            result.reason is not None
            and "Strictness fail (Conflicting Answers)" in result.reason
        )
        assert result.reason is not None and (
            ("0.5" in result.reason)
            or ("6.0" in result.reason)
            or ("14.0" in result.reason)
        )

    def test_no_penalty_ambiguity_if_gt_is_also_ambiguous(self):
        user_messages = self._get_prompt() or []
        assistant_message = Message(
            role="assistant",
            content="Solutions: $\\boxed{1}$, $\\boxed{2}$, $\\boxed{3}$.",
        )
        messages_arg = user_messages + [assistant_message]
        ground_truth_arg = "Answers: $\\boxed{1}$, $\\boxed{2}$."
        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        assert (
            result.reason is not None
            and "Strictness fail (Conflicting Answers)" in result.reason
        )
        assert (
            result.reason is not None
            and "includes other distinct numerical values not matching any original answer: [3.0]"
            in result.reason
        )

    def test_issue3_scenario_correct_handling(self):
        user_messages = self._get_prompt() or []
        ground_truth_arg = "The speed of the river is $v_R=4 \\mathrm{km/h}$, and the speed of the boat is $v_B=10 \\mathrm{km/h}$."
        response_content_str = "Let $x=\\frac{3}{v_{B}-v_{R}}, y=\\frac{7}{v_{B}+v_{R}}$. The solution to the last system is (0.5, 0.5). Then $v_R=4 \\mathrm{km} / \\mathrm{h}$, and $v_B=10 \\mathrm{km} / \\mathrm{h}$."
        assistant_message = Message(role="assistant", content=response_content_str)
        messages_arg = user_messages + [assistant_message]

        result = math_reward(messages=messages_arg, ground_truth=ground_truth_arg)
        assert isinstance(result, EvaluateResult)
        assert result.score == 0.0
        assert (
            result.reason is not None
            and "Strictness fail (Conflicting Answers)" in result.reason
        )
        assert result.reason is not None and (
            ("0.5" in result.reason)
            or ("3.0" in result.reason)
            or ("7.0" in result.reason)
        )

    def test_require_units_basic_functionality(self):
        user_messages = self._get_prompt() or []
        ground_truth_arg = "Answer is \\boxed{10 km}."

        assistant_message_match_units = Message(
            role="assistant", content="Answer is \\boxed{10 km}."
        )
        messages_match_units_arg = user_messages + [assistant_message_match_units]

        assistant_message_no_units = Message(
            role="assistant", content="Answer is \\boxed{10}."
        )
        messages_no_units_arg = user_messages + [assistant_message_no_units]

        result_match = math_reward(
            messages=messages_match_units_arg,
            ground_truth=ground_truth_arg,
            require_units=True,
        )
        assert isinstance(result_match, EvaluateResult)
        assert result_match.score == 1.0
        assert result_match.reason is None or (
            "Unit presence mismatch" not in result_match.reason
        )

        result_no_unit = math_reward(
            messages=messages_no_units_arg,
            ground_truth=ground_truth_arg,
            require_units=True,
        )
        assert isinstance(result_no_unit, EvaluateResult)
        assert result_no_unit.score == 0.0
        assert (
            result_no_unit.reason is not None
            and "Unit presence mismatch" in result_no_unit.reason
        )



import json
import os

import requests

from reward_kit.auth import get_fireworks_account_id, get_fireworks_api_key

# Get API key using the new auth module
api_key = get_fireworks_api_key()
if api_key:
    print(f"API key retrieved via auth module: {api_key[:4]}...{api_key[-4:]}")
else:
    print("No API key retrieved via auth module.")

# Get account ID using the new auth module
account_id = get_fireworks_account_id()
if account_id:
    print(f"Account ID retrieved via auth module: {account_id}")
else:
    print("No account ID retrieved via auth module.")

# Ensure api_key is not None for header construction, default to empty string if None
effective_api_key = api_key if api_key is not None else ""

# Test API connection
try:
    # Try listing models to verify API connectivity
    headers = {"Authorization": f"Bearer {effective_api_key}"}
    base_url = "https://api.fireworks.ai/v1"

    # Check if models endpoint works (to verify API connection)
    models_url = f"{base_url}/models?limit=1"
    print(f"Testing models endpoint: {models_url}")
    response = requests.get(models_url, headers=headers)
    print(f"Response: {response.status_code} - {response.reason}")
    if response.status_code == 200:
        print("Successfully connected to Fireworks API")
    else:
        print(f"Error response: {response.text}")

    if account_id:
        # Check if the evaluations endpoint is available
        eval_url = f"{base_url}/accounts/{account_id}/evaluations"
        print(f"Testing evaluations endpoint: {eval_url}")
        response = requests.get(eval_url, headers=headers)
        print(f"Response: {response.status_code} - {response.reason}")
        if response.status_code != 200:
            print(f"Error response: {response.text}")

        # Check if there's an evaluators endpoint
        evaluators_url = f"{base_url}/accounts/{account_id}/evaluators"
        print(f"Testing evaluators endpoint: {evaluators_url}")
        response = requests.get(evaluators_url, headers=headers)
        print(f"Response: {response.status_code} - {response.reason}")
        if response.status_code != 200:
            print(f"Error response: {response.text}")

        # Look for alternate endpoints
        for endpoint in ["evaluation", "evaluator"]:
            url = f"{base_url}/accounts/{account_id}/{endpoint}"
            print(f"Testing alternate endpoint: {url}")
            response = requests.get(url, headers=headers)
            print(f"Response: {response.status_code} - {response.reason}")

except Exception as e:
    print(f"Error connecting to Fireworks API: {str(e)}")
